from typing import Optional

import numpy as np
import numpy.typing as npt
import scipy as sp  # type: ignore


def empirical_covariance_upper(Y: npt.NDArray[np.double]) -> npt.NDArray[np.double]:
    """Compute the upper triangular part of the empirical covariance matrix Y
    with shape (parameters, ensemble_size).

    Examples
    --------
    >>> Y = np.array([[-2.4, -0.3,  0.7,  0.2,  1.1],
    ...               [-1.5,  0.4, -0.4, -0.9,  1. ],
    ...               [-0.1, -0.4, -0. , -0.5,  1.1]])
    >>> empirical_covariance_upper(Y)
    array([[1.873, 0.981, 0.371],
           [0.   , 0.997, 0.392],
           [0.   , 0.   , 0.407]])

    Naive computation:

    >>> empirical_cross_covariance(Y, Y)
    array([[1.873, 0.981, 0.371],
           [0.981, 0.997, 0.392],
           [0.371, 0.392, 0.407]])
    """
    _, num_observations = Y.shape
    Y = (Y - np.mean(Y, axis=1, keepdims=True)) / np.sqrt(num_observations - 1)
    # https://www.math.utah.edu/software/lapack/lapack-blas/dsyrk.html
    YYT: npt.NDArray[np.double] = sp.linalg.blas.dsyrk(alpha=1.0, a=Y)
    return YYT


def empirical_cross_covariance(
    X: npt.NDArray[np.double], Y: npt.NDArray[np.double]
) -> npt.NDArray[np.double]:
    """Both X and Y have shape (parameters, ensemble_size).

    We use this function instead of np.cov to handle cross-correlation,
    where X and Y have a different number of parameters.

    Examples
    --------
    >>> X = np.array([[-2.4, -0.3,  0.7],
    ...               [ 0.2,  1.1, -1.5]])
    >>> Y = np.array([[ 0.4, -0.4, -0.9],
    ...               [ 1. , -0.1, -0.4],
    ...               [-0. , -0.5,  1.1],
    ...               [-1.8, -1.1,  0.3]])
    >>> empirical_cross_covariance(X, Y)
    array([[-1.035     , -1.15833333,  0.66      ,  1.56333333],
           [ 0.465     ,  0.36166667, -1.08      , -1.09666667]])

    Verify against numpy.cov

    >>> np.cov(X, rowvar=True, ddof=1)
    array([[ 2.50333333, -0.99666667],
           [-0.99666667,  1.74333333]])
    >>> empirical_cross_covariance(X, X)
    array([[ 2.50333333, -0.99666667],
           [-0.99666667,  1.74333333]])

    """
    assert X.shape[1] == Y.shape[1], "Ensemble size must be equal"

    # https://en.wikipedia.org/wiki/Estimation_of_covariance_matrices
    # Subtract mean. Even though the equation says E[(X - E[X])(Y - E[Y])^T],
    # we actually only need to subtract the mean value from one matrix, since
    # (X - E[X])(Y - E[Y])^T = E[(X - E[X])Y] - E[(X - E[X])E[Y]^T]
    # = E[(X - E[X])Y] - E[(0)E[Y]^T] = E[(X - E[X])Y]
    # We choose to subtract from the matrix with the smaller number of rows
    if X.shape[0] > Y.shape[0]:
        Y = Y - np.mean(Y, axis=1, keepdims=True)
    else:
        X = X - np.mean(X, axis=1, keepdims=True)

    # Compute outer product and divide
    # If X is a large matrix, it might be stored as a float32 array to save memory.
    # However, if Y is of type float64,
    # the resulting cross-covariance matrix will be float64,
    # potentially doubling the memory usage even if X is float32.
    # To prevent unnecessary memory consumption,
    # we cast Y to the same data type as X before computing the dot product.
    # This ensures that the output cross-covariance matrix uses memory efficiently
    # while retaining the precision dictated by X's data type.
    cov = X @ Y.astype(X.dtype).T / (X.shape[1] - 1)
    assert cov.shape == (X.shape[0], Y.shape[0])
    return cov


def normalize_alpha(alpha: npt.NDArray[np.double]) -> npt.NDArray[np.double]:
    """Assure that sum_i (1/alpha_i) = 1.

    This is Eqn (22) in :cite:t:`EMERICK2013`.

    Examples
    --------
    >>> alpha = np.arange(10) + 1
    >>> float(np.sum(1/normalize_alpha(alpha)))
    1.0
    """
    factor = np.sum(1 / alpha)
    rescaled: npt.NDArray[np.double] = alpha * factor
    return rescaled


def singular_values_to_keep(
    singular_values: npt.NDArray[np.double], truncation: float = 1.0
) -> int:
    """Find the index of the singular values to keep when truncating.

    Examples
    --------
    >>> singular_values = np.array([3, 2, 1, 0, 0, 0])
    >>> i = singular_values_to_keep(singular_values, truncation=1.0)
    >>> singular_values[:i]
    array([3, 2, 1])

    >>> singular_values = np.array([4, 3, 2, 1])
    >>> i = singular_values_to_keep(singular_values, truncation=1.0)
    >>> singular_values[:i]
    array([4, 3, 2, 1])

    >>> singular_values = np.array([4, 3, 2, 1])
    >>> singular_values_to_keep(singular_values, truncation=0.95)
    4
    >>> singular_values_to_keep(singular_values, truncation=0.9)
    3
    >>> singular_values_to_keep(singular_values, truncation=0.7)
    2

    """
    assert np.all(np.diff(singular_values) <= 0), (
        "Singular values must be sorted decreasing"
    )
    assert 0 < truncation <= 1, "Threshold must be in range (0, 1]"
    singular_values = np.array(singular_values, dtype=float)

    # Take cumulative sum and normalize
    cumsum = np.cumsum(singular_values)
    cumsum /= cumsum[-1]
    return int(np.searchsorted(cumsum, v=truncation, side="left") + 1)


# =============================================================================
# INVERSION FUNCTIONS
# =============================================================================
# All of these functions compute (exactly, or approximately), the product
#
#  C_MD @ inv(C_DD + alpha * C_D) @ (D - Y)
#
# where C_MD = empirical_cross_covariance(X, Y) = center(X) @ center(Y).T
#               / (X.shape[1] - 1)
#       C_DD = empirical_cross_covariance(Y, Y) = center(Y) @ center(Y).T
#               / (Y.shape[1] - 1)
#
# The methods can be classified as
#   - exact : with truncation=1.0, these methods compute the exact solution
#   - exact : with truncation<1.0, these methods may approximate the solution
#   - approximate: if ensemble_members <= num_outputs, then the solution is
#                  always approximated, regardless of the truncation
#   - approximate: if ensemble_members > num_outputs, then the solution is
#                  exact when truncation is 1.0

# Every inversion function has the form
# inversion_<exact/approximate>_<name>


def inversion_exact_naive(
    *,
    alpha: float,
    C_D: npt.NDArray[np.double],
    D: npt.NDArray[np.double],
    Y: npt.NDArray[np.double],
    X: npt.NDArray[np.double],
    truncation: float = 1.0,
) -> npt.NDArray[np.double]:
    """Naive inversion, used for testing only.

    Computes C_MD @ inv(C_DD + alpha * C_D) @ (D - Y) naively.
    """
    # Naive implementation of Equation (3) in Emerick (2013)
    C_MD = empirical_cross_covariance(X, Y)
    C_DD = empirical_cross_covariance(Y, Y)
    C_D = np.diag(C_D) if C_D.ndim == 1 else C_D
    return C_MD @ sp.linalg.inv(C_DD + alpha * C_D) @ (D - Y)  # type: ignore


def inversion_exact_cholesky(
    *,
    alpha: float,
    C_D: npt.NDArray[np.double],
    D: npt.NDArray[np.double],
    Y: npt.NDArray[np.double],
    X: Optional[npt.NDArray[np.double]],
    truncation: float = 1.0,
    return_T: bool = False,
) -> npt.NDArray[np.double]:
    """Computes an exact inversion using `sp.linalg.solve`, which uses a
    Cholesky factorization in the case of symmetric, positive definite matrices.

    The goal is to compute: C_MD @ inv(C_DD + alpha * C_D) @ (D - Y)

    First we solve (C_DD + alpha * C_D) @ T = (D - Y) for T, so that
    T = inv(C_DD + alpha * C_D) @ (D - Y), then we compute
    C_MD @ T, but we don't explicitly form C_MD, since it might be more
    efficient to perform the matrix products in another order.
    """
    C_DD = empirical_covariance_upper(Y)  # Only compute upper part

    # Arguments for sp.linalg.solve
    solver_kwargs = {
        "overwrite_a": True,
        "overwrite_b": True,
        "assume_a": "pos",  # Assume positive definite matrix (use cholesky)
        "lower": False,  # Only use the upper part while solving
    }

    # Compute T := sp.linalg.inv(C_DD + alpha * C_D) @ (D - Y)
    if C_D.ndim == 2:
        # C_D is a covariance matrix
        C_DD += alpha * C_D  # Save memory by mutating
        T = sp.linalg.solve(C_DD, D - Y, **solver_kwargs)
    elif C_D.ndim == 1:
        # C_D is an array, so add it to the diagonal without forming diag(C_D)
        C_DD.flat[:: C_DD.shape[1] + 1] += alpha * C_D
        T = sp.linalg.solve(C_DD, D - Y, **solver_kwargs)

    # Center matrix
    Y = Y - np.mean(Y, axis=1, keepdims=True)
    _, num_ensemble = Y.shape

    # Don't left-multiply the X
    if return_T:
        return (Y.T @ T) / (num_ensemble - 1)  # type: ignore

    return np.linalg.multi_dot([X, Y.T / (num_ensemble - 1), T])  # type: ignore


def inversion_exact_lstsq(
    *,
    alpha: float,
    C_D: npt.NDArray[np.double],
    D: npt.NDArray[np.double],
    Y: npt.NDArray[np.double],
    X: npt.NDArray[np.double],
    truncation: float = 1.0,
) -> npt.NDArray[np.double]:
    """Computes inversion using least squares. While this method can deal with
    rank-deficient C_D, it should not be used since it's very slow.
    """
    C_DD = empirical_cross_covariance(Y, Y)

    # A covariance matrix was given
    if C_D.ndim == 2:
        lhs = C_DD + alpha * C_D
    # A diagonal covariance matrix was given as a vector
    else:
        lhs = C_DD
        lhs.flat[:: lhs.shape[0] + 1] += alpha * C_D

    # T = lhs^-1 @ (D - Y)
    # lhs @ T = (D - Y)
    ans, *_ = sp.linalg.lstsq(
        lhs, D - Y, overwrite_a=True, overwrite_b=True, lapack_driver="gelsy"
    )

    # Compute C_MD := X @ center(Y).T / (Y.shape[1] - 1)
    Y_shift = (Y - np.mean(Y, axis=1, keepdims=True)) / (Y.shape[1] - 1)
    return np.linalg.multi_dot([X, Y_shift.T, ans])  # type: ignore


def inversion_exact_rescaled(
    *,
    alpha: float,
    C_D: npt.NDArray[np.double],
    D: npt.NDArray[np.double],
    Y: npt.NDArray[np.double],
    X: npt.NDArray[np.double],
    truncation: float = 1.0,
) -> npt.NDArray[np.double]:
    """Compute a rescaled inversion.

    See Appendix A.1 in :cite:t:`emerickHistoryMatchingTimelapse2012`
    for details regarding this approach.
    """
    C_DD = empirical_cross_covariance(Y, Y)

    if C_D.ndim == 2:
        # Eqn (57). Cholesky factorize the covariance matrix C_D
        # TODO: here we compute the cholesky factor in every call, but C_D
        # never changes. it would be better to compute it once
        C_D_L = sp.linalg.cholesky(C_D, lower=True)  # Lower triangular cholesky
        C_D_L_inv, _ = sp.linalg.lapack.dtrtri(
            C_D_L, lower=1, overwrite_c=1
        )  # Invert lower triangular using BLAS routine
        C_D_L_inv /= np.sqrt(alpha)

        # Eqn (59). Form C_tilde
        # TODO: Use BLAS routine for triangular times dense matrix
        # sp.linalg.blas.strmm(alpha=1, a=C_D_L_inv, b=C_DD, lower=1)
        C_tilde = C_D_L_inv @ C_DD @ C_D_L_inv.T
        C_tilde.flat[:: C_tilde.shape[0] + 1] += 1.0  # Add to diagonal

    # When C_D is a diagonal covariance matrix, there is no need to perform
    # the cholesky factorization
    elif C_D.ndim == 1:
        C_D_L_inv = 1 / np.sqrt(C_D * alpha)
        C_tilde = (C_D_L_inv * (C_DD * C_D_L_inv).T).T
        C_tilde.flat[:: C_tilde.shape[0] + 1] += 1.0  # Add to diagonal

    # Eqn (60). Compute SVD, which is equivalent to taking eigendecomposition
    # since C_tilde is PSD. Using eigh() is faster than svd().
    # Note that svd() returns eigenvalues in decreasing order, while eigh()
    # returns eigenvalues in increasing order.
    # driver="evr" => fastest option
    s, U = sp.linalg.eigh(C_tilde, driver="evr", overwrite_a=True)
    # Truncate the SVD ( U_r @ np.diag(s_r) @ U_r.T == C_tilde )
    # N_n is the number of observations
    # N_e is the number of members in the ensemble
    N_n, N_e = Y.shape

    idx = singular_values_to_keep(s[::-1], truncation=truncation)
    N_r = min(N_n, N_e - 1, idx)  # Number of values in SVD to keep
    U_r, s_r = U[:, -N_r:], s[-N_r:]
    # U_r @ np.diag(s_r) @ U_r.T == C_tilde

    # Eqn (61). Compute symmetric term once first, then multiply together and
    # finally multiply with (D - Y)
    term = C_D_L_inv.T @ U_r if C_D.ndim == 2 else (C_D_L_inv * U_r.T).T

    # Compute the first factors, which make up C_MD
    Y_shift = (Y - np.mean(Y, axis=1, keepdims=True)) / (N_e - 1)

    return np.linalg.multi_dot(  # type: ignore
        [X, Y_shift.T, term / s_r, term.T, (D - Y)]
    )


def inversion_exact_subspace_woodbury(
    *,
    alpha: float,
    C_D: npt.NDArray[np.double],
    D: npt.NDArray[np.double],
    Y: npt.NDArray[np.double],
    X: npt.NDArray[np.double],
    truncation: float = 1.0,
) -> npt.NDArray[np.double]:
    """Use the Woodbury lemma to compute the inversion.

    This approach uses the Woodbury lemma to compute:
        C_MD @ inv(C_DD + alpha * C_D) @ (D - Y)

    Since C_DD = U @ U.T, where U := center(Y) / sqrt(N_e - 1), we can use:

    (A + U @ U.T)^-1 = A^-1 - A^-1 @ U @ (1 + U.T @ A^-1 @ U )^-1 @ U.T @ A^-1

    to compute inv(C_DD + alpha * C_D).
    """

    # Woodbury:
    # (A + U @ U.T)^-1 = A^-1 - A^-1 @ U @ (1 + U.T @ A^-1 @ U )^-1 @ U.T @ A^-1

    # Compute D_delta. N_n = number of outputs, N_e = number of ensemble members
    N_n, N_e = Y.shape
    D_delta = Y - np.mean(Y, axis=1, keepdims=True)  # Subtract average
    D_delta /= np.sqrt(N_e - 1)

    # Compute the first factors, which make up C_MD
    # X_shift = (X - np.mean(X, axis=1, keepdims=True)) / np.sqrt(N_e - 1)

    # A full covariance matrix was given
    if C_D.ndim == 2:
        # Invert C_D
        # TODO: This inverse could be cached
        C_D_inv = np.linalg.inv(C_D) / alpha

        # Compute the center part of the rhs in woodburry
        center = np.linalg.multi_dot([D_delta.T, C_D_inv, D_delta])
        center.flat[:: center.shape[0] + 1] += 1.0  # Add to diagonal

        # Compute the symmetric term of the rhs in woodbury
        term = C_D_inv @ D_delta

        # Compute the woodbury inversion, then return
        inverted = C_D_inv - np.linalg.multi_dot([term, sp.linalg.inv(center), term.T])
        return np.linalg.multi_dot(  # type: ignore
            [X, D_delta.T / np.sqrt(N_e - 1), inverted, (D - Y)]
        )

    # A diagonal covariance matrix was given as a 1D array.
    # Same computation as above, but exploit the diagonal structure
    else:
        C_D_inv = 1 / (C_D * alpha)  # Invert diagonal
        center = np.linalg.multi_dot([D_delta.T * C_D_inv, D_delta])
        center.flat[:: center.shape[0] + 1] += 1.0
        UT_D = D_delta.T * C_D_inv
        inverted = np.diag(C_D_inv) - np.linalg.multi_dot(
            [UT_D.T, sp.linalg.inv(center), UT_D]
        )
        return np.linalg.multi_dot(  # type: ignore
            [X, D_delta.T / np.sqrt(N_e - 1), inverted, (D - Y)]
        )


def inversion_subspace(
    *,
    alpha: float,
    C_D: npt.NDArray[np.double],
    D: npt.NDArray[np.double],
    Y: npt.NDArray[np.double],
    X: Optional[npt.NDArray[np.double]],
    truncation: float = 1.0,
    return_T: bool = False,
) -> npt.NDArray[np.double]:
    """See Appendix A.2 in :cite:t:`emerickHistoryMatchingTimelapse2012`.

    This is an approximate solution. The approximation is that when
    U, w, V.T = svd(D_delta)
    then we assume that U @ U.T = I.
    This is not true in general, for instance:

    >>> Y = np.array([[2, 0],
    ...               [0, 0],
    ...               [0, 0]])
    >>> D_delta = Y - np.mean(Y, axis=1, keepdims=True)  # Subtract average
    >>> D_delta
    array([[ 1., -1.],
           [ 0.,  0.],
           [ 0.,  0.]])
    >>> U, w, VT = sp.linalg.svd(D_delta)
    >>> U, w
    (array([[1., 0., 0.],
           [0., 1., 0.],
           [0., 0., 1.]]), array([1.41421356, 0.        ]))
    >>> U[:, :1] @ np.diag(w[:1]) @ VT[:1, :] # Reconstruct D_Delta
    array([[ 1., -1.],
           [ 0.,  0.],
           [ 0.,  0.]])
    >>> U[:, :1] @ U[:, :1].T # But U_r @ U_r.T != I
    array([[1., 0., 0.],
           [0., 0., 0.],
           [0., 0., 0.]])

    """

    # N_n is the number of observations
    # N_e is the number of members in the ensemble
    N_n, N_e = Y.shape

    # Subtract the mean of every observation, see Eqn (67)
    D_delta = Y - np.mean(Y, axis=1, keepdims=True)  # Subtract average

    # Eqn (68)
    # TODO: Approximately 50% of the time in the function is spent here
    # consider using randomized svd for further speed gains
    U, w, _ = sp.linalg.svd(D_delta, full_matrices=False)

    # Clip the singular value decomposition
    idx = singular_values_to_keep(w, truncation=truncation)
    N_r = min(N_n, N_e - 1, idx)  # Number of values in SVD to keep
    U_r, w_r = U[:, :N_r], w[:N_r]

    # Eqn (70). First compute the symmetric term, then form X
    U_r_w_inv = U_r / w_r
    if C_D.ndim == 1:
        X1 = (N_e - 1) * np.linalg.multi_dot([U_r_w_inv.T * C_D * alpha, U_r_w_inv])
    else:
        X1 = (N_e - 1) * np.linalg.multi_dot([U_r_w_inv.T, alpha * C_D, U_r_w_inv])

    # Eqn (72)
    # Z, T, _ = sp.linalg.svd(X, overwrite_a=True, full_matrices=False)
    # Compute SVD, which is equivalent to taking eigendecomposition
    # since X is PSD. Using eigh() is faster than svd().
    # Note that svd() returns eigenvalues in decreasing order, while eigh()
    # returns eigenvalues in increasing order.
    # driver="evr" => fastest option
    T, Z = sp.linalg.eigh(X1, driver="evr", overwrite_a=True)

    # Eqn (74).
    # C^+ = (N_e - 1) hat{C}^+
    #     = (N_e - 1) (U / w @ Z) * (1 / (1 + T)) (U / w @ Z)^T
    #     = (N_e - 1) (term) * (1 / (1 + T)) (term)^T
    # and finally we multiiply by (D - Y)
    term = U_r_w_inv @ Z

    if return_T:
        return np.linalg.multi_dot(  # type: ignore
            [D_delta.T, (term / (1 + T)), term.T, (D - Y)]
        )

    # Compute C_MD = center(X) @ center(Y).T / (num_ensemble - 1)
    return np.linalg.multi_dot(  # type: ignore
        [X, D_delta.T, (term / (1 + T)), term.T, (D - Y)]
    )


def inversion_rescaled_subspace(
    *,
    alpha: float,
    C_D: npt.NDArray[np.double],
    D: npt.NDArray[np.double],
    Y: npt.NDArray[np.double],
    X: npt.NDArray[np.double],
    truncation: float = 1.0,
) -> npt.NDArray[np.double]:
    """
    See Appendix A.2 in :cite:t:`emerickHistoryMatchingTimelapse2012`.

    Subspace inversion with rescaling.
    """
    # TODO: I don't understand why this approach is not approximate, when
    # `inversion_subspace` is approximate

    N_n, N_e = Y.shape
    D_delta = Y - np.mean(Y, axis=1, keepdims=True)  # Subtract average

    if C_D.ndim == 2:
        # Eqn (76). Cholesky factorize the covariance matrix C_D
        # TODO: here we compute the cholesky factor in every call, but C_D
        # never changes. it would be better to compute it once
        C_D_L = sp.linalg.cholesky(C_D * alpha, lower=True)  # Lower triangular cholesky
        # Here C_D_L is C^{1/2} in equation (57)
        # assert np.allclose(C_D_L @ C_D_L.T, C_D * alpha)
        C_D_L_inv, _ = sp.linalg.lapack.dtrtri(
            C_D_L, lower=1
        )  # Invert lower triangular

        # Use BLAS to compute product of lower triangular matrix C_D_L_inv and D_Delta
        # This line is equal to C_D_L_inv @ D_delta
        C_D_L_times_D_delta = sp.linalg.blas.dtrmm(
            alpha=1.0, a=C_D_L_inv, b=D_delta, lower=1
        )

    else:
        # Same as above, but C_D is a vector
        C_D_L_inv = 1 / np.sqrt(alpha * C_D)  # Invert the Cholesky factor a diagonal
        C_D_L_times_D_delta = (D_delta.T * C_D_L_inv).T

    U, w, _ = sp.linalg.svd(C_D_L_times_D_delta, overwrite_a=True, full_matrices=False)
    idx = singular_values_to_keep(w, truncation=truncation)

    # assert np.allclose(VT @ VT.T, np.eye(VT.shape[0]))
    N_r = min(N_n, N_e - 1, idx)  # Number of values in SVD to keep
    U_r, w_r = U[:, :N_r], w[:N_r]

    # Eqn (78) - taking into account that C_D_L_inv could be an array
    term = C_D_L_inv.T @ (U_r / w_r) if C_D.ndim == 2 else ((U_r / w_r).T * C_D_L_inv).T
    T_r = (N_e - 1) / w_r**2  # Equation (79)
    diag = 1 / (1 + T_r)

    # Compute C_MD
    return np.linalg.multi_dot(  # type: ignore
        [X, D_delta.T, (term * diag), term.T, (D - Y)]
    )


if __name__ == "__main__":
    import pytest

    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(
        args=[
            __file__,
            "--doctest-modules",
            "-v",
            "-v",
            # "-k simple",
        ]
    )



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.15.2
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %%
# ruff: noqa: E402

# %% [markdown]
# # Linear regression with ESMDA
#
# We solve a linear regression problem using ESMDA.
# First we define the forward model as $g(x) = Ax$,
# then we set up a prior ensemble on the linear
# regression coefficients, so $x \sim \mathcal{N}(0, 1)$.
#
# As shown in the 2013 paper by Emerick et al, when a set of
# inflation weights $\alpha_i$ is chosen so that $\sum_i \alpha_i^{-1} = 1$,
# ESMDA yields the correct posterior mean for the linear-Gaussian case.

# %% [markdown]
# ## Import packages

# %%
import numpy as np
from matplotlib import pyplot as plt

from iterative_ensemble_smoother import ESMDA

# %% [markdown]
# ## Create problem data
#
# Some settings worth experimenting with:
#
# - Decreasing `prior_std=1` will pull the posterior solution toward zero.
# - Increasing `num_ensemble` will increase the quality of the solution.
# - Increasing `num_observations / num_parameters`
#   will increase the quality of the solution.

# %%
num_parameters = 25
num_observations = 100
num_ensemble = 30
prior_std = 1

# %%
rng = np.random.default_rng(42)

# Create a problem with g(x) = A @ x
A = rng.standard_normal(size=(num_observations, num_parameters))


def g(X):
    """Forward model."""
    return A @ X


# Create observations: obs = g(x) + N(0, 1)
x_true = np.linspace(-1, 1, num=num_parameters)
observation_noise = rng.standard_normal(size=num_observations)
observations = g(x_true) + observation_noise

# Initial ensemble X ~ N(0, prior_std) and diagonal covariance with ones
X = rng.normal(size=(num_parameters, num_ensemble)) * prior_std

# Covariance matches the noise added to observations above
covariance = np.ones(num_observations)

# %% [markdown]
# ## Solve the maximum likelihood problem
#
# We can solve $Ax = b$, where $b$ is the observations,
# for the maximum likelihood estimate.
# Notice that unlike using a Ridge model,
# solving $Ax = b$ directly does not use any prior information.

# %%
x_ml, *_ = np.linalg.lstsq(A, observations, rcond=None)

plt.figure(figsize=(8, 3))
plt.scatter(np.arange(len(x_true)), x_true, label="True parameter values")
plt.scatter(np.arange(len(x_true)), x_ml, label="ML estimate (no prior)")
plt.xlabel("Parameter index")
plt.ylabel("Parameter value")
plt.grid(True, ls="--", zorder=0, alpha=0.33)
plt.legend()
plt.show()

# %% [markdown]
# ## Solve using ESMDA
#
# We crease an `ESMDA` instance and solve the Guass-linear problem.

# %%
smoother = ESMDA(
    covariance=covariance,
    observations=observations,
    alpha=5,
    seed=1,
)

X_i = np.copy(X)
for i, alpha_i in enumerate(smoother.alpha, 1):
    print(
        f"ESMDA iteration {i}/{smoother.num_assimilations()}"
        + f" with inflation factor alpha_i={alpha_i}"
    )
    X_i = smoother.assimilate(X_i, Y=g(X_i))


X_posterior = np.copy(X_i)

# %% [markdown]
# ## Plot and compare solutions
#
# Compare the true parameters with both the ML estimate
# from linear regression and the posterior means obtained using `ESMDA`.

# %%
plt.figure(figsize=(8, 3))
plt.scatter(np.arange(len(x_true)), x_true, label="True parameter values")
plt.scatter(np.arange(len(x_true)), x_ml, label="ML estimate (no prior)")
plt.scatter(
    np.arange(len(x_true)), np.mean(X_posterior, axis=1), label="Posterior mean"
)
plt.xlabel("Parameter index")
plt.ylabel("Parameter value")
plt.grid(True, ls="--", zorder=0, alpha=0.33)
plt.legend()
plt.show()

# %% [markdown]
# We now include the posterior samples as well.

# %%
plt.figure(figsize=(8, 3))
plt.scatter(np.arange(len(x_true)), x_true, label="True parameter values")
plt.scatter(np.arange(len(x_true)), x_ml, label="ML estimate (no prior)")
plt.scatter(
    np.arange(len(x_true)), np.mean(X_posterior, axis=1), label="Posterior mean"
)

# Loop over every ensemble member and plot it
for j in range(num_ensemble):
    # Jitter along the x-axis a little bit
    x_jitter = np.arange(len(x_true)) + rng.normal(loc=0, scale=0.1, size=len(x_true))

    # Plot this ensemble member
    plt.scatter(
        x_jitter,
        X_posterior[:, j],
        label=("Posterior values" if j == 0 else None),
        color="black",
        alpha=0.2,
        s=5,
        zorder=0,
    )
plt.xlabel("Parameter index")
plt.ylabel("Parameter value")
plt.grid(True, ls="--", zorder=0, alpha=0.33)
plt.legend()
plt.show()



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

import numpy as np
import scipy as sp  # type: ignore

if TYPE_CHECKING:
    import numpy.typing as npt

from iterative_ensemble_smoother.sies_inversion import (
    inversion_direct,
    inversion_subspace_exact,
    inversion_subspace_projected,
)
from iterative_ensemble_smoother.utils import _validate_inputs, sample_mvnormal


class SIES:
    r"""
    Implement a Sequential Iterative Ensemble Smoother (SIES) for Data Assimilation.

    This is an implementation of the algorithm described in the paper: "Efficient
    Implementation of an Iterative Ensemble Smoother for Data Assimilation
    and Reservoir History Matching", written by :cite:t:`evensen2019efficient`.

    """

    inversion_funcs = {
        "direct": inversion_direct,
        "subspace_exact": inversion_subspace_exact,
        "subspace_projected": inversion_subspace_projected,
    }

    def __init__(
        self,
        parameters: npt.NDArray[np.double],
        covariance: npt.NDArray[np.double],
        observations: npt.NDArray[np.double],
        *,
        inversion: str = "subspace_exact",
        truncation: float = 1.0,
        seed: Union[None, int, np.random._generator.Generator] = None,
    ) -> None:
        """
        Initialize the instance.

        Parameters
        ----------
        parameters : npt.NDArray[np.double]
            A 2D array of shape (num_parameters, ensemble_size). Each row corresponds
            to a parameter in the model, and each column corresponds to an ensemble
            member (realization). This is X in Evensen (2019).
        covariance : npt.NDArray[np.double]
            Either a 1D array of diagonal covariances, or a 2D covariance matrix.
            The shape is either (num_observations,) or (num_observations,
            num_observations).
            This is C_dd in Evensen (2019), and represents observation or measurement
            errors. We observe d from the real world, y from the model g(x), and
            assume that d = y + e, where the error e is multivariate normal with
            covariance given by `covariance`.
        observations : npt.NDArray[np.double]
            A 1D array of observations, with shape (num_observations,).
            This is d in Evensen (2019).
        inversion : str
            The type of inversion used in the algorithm. Every inversion method
            scales the variables. The default is `subspace.`
            The options are:

                * `direct`:
                    Solve Eqn (42) directly, which involves inverting a
                    matrix of shape (num_observations, num_observations).
                * `subspace_exact` :
                    Solve Eqn (42) using Eqn (50), i.e., the Woodbury
                    lemma to invert a matrix of size (ensemble_size, ensemble_size).
                * `subspace_projected` :
                    Solve Eqn (42) using Section 3.3, i.e., by projecting the covariance
                    onto S. This approach utilizes the truncation factor `truncation`.

        truncation : float
            How much of the total energy (singular values squared) to keep in the
            SVD when `inversion` equals `subspace_projected`. Choosing 1.0
            retains all information, while 0.0 removes all information.
            The default is 1.0.
        seed : Union[None, int, np.random._generator.Generator], optional
            Integer used to seed the random number generator. The default is None.
        """
        _validate_inputs(
            parameters=parameters, covariance=covariance, observations=observations
        )

        self.rng = np.random.default_rng(seed)
        self.inversion = self.inversion_funcs[inversion]
        self.truncation = truncation
        self.X = parameters
        self.d = observations
        self.C_dd = covariance
        self.A = (self.X - self.X.mean(axis=1, keepdims=True)) / np.sqrt(
            self.X.shape[1] - 1
        )

        # Create and store the cholesky factorization of C_dd
        if self.C_dd.ndim == 2:
            # With lower=True, the cholesky factorization obeys:
            # C_dd_cholesky @ C_dd_cholesky.T = C_dd
            self.C_dd_cholesky = sp.linalg.cholesky(
                self.C_dd,
                lower=True,
                overwrite_a=False,
                check_finite=True,
            )
        else:
            self.C_dd_cholesky = np.sqrt(self.C_dd)

        # Equation (14)
        self.D = (
            self.d.reshape(-1, 1)
            + sample_mvnormal(
                C_dd_cholesky=self.C_dd_cholesky, rng=self.rng, size=self.X.shape[1]
            )
        ).astype(parameters.dtype, copy=False)

        self.W = np.zeros(
            shape=(self.X.shape[1], self.X.shape[1]), dtype=parameters.dtype
        )

    def evaluate_objective(
        self, W: npt.NDArray[np.double], Y: npt.NDArray[np.double]
    ) -> npt.NDArray[np.double]:
        """Evaluate the objective function in Equation (18), taking the mean
        over each ensemble member.

        Parameters
        ----------
        W : npt.NDArray[np.double]
            Current weight matrix W, which represents the current X_i as a linear
            combination of the prior. See equation (17) and line 9 in Algorithm 1.
        Y : npt.NDArray[np.double]
            Responses when evaluating the model at X_i. In other words, Y = g(X_i).

        Returns
        -------
        float
            The objective function value. Lower objective is better.

        """
        (prior, likelihood) = self.evaluate_objective_elementwise(W, Y)
        answer: npt.NDArray[np.double] = (prior + likelihood).mean()
        return answer

    def evaluate_objective_elementwise(
        self, W: npt.NDArray[np.double], Y: npt.NDArray[np.double]
    ) -> tuple[npt.NDArray[np.double], npt.NDArray[np.double]]:
        """Equation (18) elementwise and termwise, returning a tuple of vectors
        (prior, likelihood).

        Parameters
        ----------
        W : npt.NDArray[np.double]
            Current weight matrix W, which represents the current X_i as a linear
            combination of the prior. See equation (17) and line 9 in Algorithm 1.
        Y : npt.NDArray[np.double]
            Responses when evaluating the model at X_i. In other words, Y = g(X_i).

        Returns
        -------
        prior : npt.NDArray[np.double]
            A 1D array representing distance from prior for each ensemble member.
        likelihood : npt.NDArray[np.double]
            A 1D array representing distance from observations for each ensemble member.

        """
        # Evaluate the elementwise prior term
        # Evaluate np.diag(W.T @ W) = (W**2).sum(axis=0)
        prior = (W**2).sum(axis=0)

        # Introduce F as the differences between responses y=g(x) and D
        F = Y - self.D

        if self.C_dd.ndim == 2:
            # Evaluate the elementwise likelihood term
            # Evaluate np.diag((g(X_i) - D).T @ C_dd^{-1} @ (g(X_i) - D))
            #        = np.diag((Y - D).T @ C_dd^{-1} @ (Y - D))
            # First let F := Y - D, then use the cholesky factors C_dd = L @ L.T
            # to write (Y - D).T @ C_dd^{-1} @ (Y - D)
            #             F.T @ (L @ L.T)^-1 @ F
            #             (L^-1 @ F).T @ (L^-1 @ F)
            # To compute the expression above, we solve K := L^-1 @ F,
            # then we compute np.diag(K.T @ K) as (K**2).sum(axis=0)
            # Likelihood is equal to: np.diag(F.T @ inv(self.C_dd) @ F)
            K = sp.linalg.solve_triangular(a=self.C_dd_cholesky, b=F, lower=True)
            likelihood = (K**2).sum(axis=0)

        else:
            # If C_dd is diagonal, then (L^-1 @ F) = F / L.reshape(-1, 1)
            K = F / self.C_dd_cholesky.reshape(-1, 1)
            likelihood = (K**2).sum(axis=0)

        return (prior, likelihood)

    def sies_iteration(
        self,
        responses: npt.NDArray[np.double],
        step_length: float = 0.5,
        ensemble_mask: Optional[npt.NDArray[np.bool_]] = None,
    ) -> npt.NDArray[np.double]:
        """Perform a single SIES iteration (Gauss-Newton step).

        This method implements lines 4-9 in Algorithm 1.
        It returns an updated X and updates the internal state W.

        Parameters
        ----------
        responses : npt.NDArray[np.double]
            The model evaluated at X_i. In other words, responses = g(X_i).
            This is Y in the paper.
        step_length : float, optional
            Step length for Gauss-Newton. The default is 0.5.

        Returns
        -------
        npt.NDArray[np.double]
            Updated parameter ensemble.

        """
        Y = responses

        if ensemble_mask is not None:
            # Lines 4 through 8
            proposed_W = self.propose_W_masked(
                Y, ensemble_mask=ensemble_mask, step_length=step_length
            )
            self.W[:, ensemble_mask] = proposed_W

        else:
            # Lines 4 through 8
            proposed_W = self.propose_W(Y, step_length=step_length)
            self.W = proposed_W

        # Line 9
        N = self.X.shape[1]  # Ensemble members
        ans: npt.NDArray[np.double] = self.X + self.X @ self.W / np.sqrt(N - 1)
        return ans.astype(responses.dtype)

    def propose_W(
        self, responses: npt.NDArray[np.double], step_length: float = 0.5
    ) -> npt.NDArray[np.double]:
        """Return a proposal for W_i, without updating the internal W.

        This is an implementation of lines 4-8 in Algorithm 1.

        Parameters
        ----------
        responses : npt.NDArray[np.double]
            The model evaluated at X_i. In other words, responses = g(X_i).
            This is Y in the paper.
        step_length : float, optional
            Step length for Gauss-Newton. The default is 0.5.

        Returns
        -------
        W_i : npt.NDArray[np.double]
            A proposal for a new W in the algorithm.

        """
        Y = responses
        assert Y.ndim == 2
        assert Y.shape[0] == self.C_dd.shape[0]
        g_X = Y.copy()

        # Get shapes. Same notation as used in the paper.
        N = self.X.shape[1]  # Ensemble members
        n = self.X.shape[0]  # Parameters (inputs)
        m = self.C_dd.shape[0]  # Responses (outputs)

        # Line 4 in Algorithm 1
        Y = (g_X - g_X.mean(axis=1, keepdims=True)) / np.sqrt(N - 1)

        # Line 5
        Omega = self.W.copy()
        Omega -= Omega.mean(axis=1, keepdims=True)
        Omega /= np.sqrt(N - 1)
        np.fill_diagonal(Omega, Omega.diagonal() + 1)  # Add identity in place

        # Remind ourselves of shapes
        assert Omega.shape == (N, N)
        assert Y.shape == (m, N)

        # Line 6
        if n >= N - 1:  # Section (2.4.1)
            S = sp.linalg.solve(Omega.T, Y.T).T

        else:  # Section (2.4.3)
            # An alternative approach to producing A_i would be keeping the
            # returned value from the previous Newton iteration (call it X_i),
            # then computing:
            # A_i = (self.X_i - self.X_i.mean(axis=1, keepdims=True)) / np.sqrt(N - 1)
            A_i = self.A @ Omega
            S = sp.linalg.solve(
                Omega.T, np.linalg.multi_dot([Y, sp.linalg.pinv(A_i), A_i]).T
            ).T

        # Line 7
        H = S @ self.W + self.D - g_X

        # Line 8
        # Some asserts to remind us what the shapes of the matrices are
        assert self.W.shape == (N, N)
        assert S.shape == (m, N)
        assert self.C_dd.shape in [(m, m), (m,)]
        assert H.shape == (m, N)

        proposed_W: npt.NDArray[np.double] = self.inversion(
            W=self.W,
            step_length=step_length,
            S=S,
            C_dd=self.C_dd,
            H=H,
            C_dd_cholesky=self.C_dd_cholesky,
            truncation=self.truncation,
        )
        return proposed_W

    def propose_W_masked(
        self,
        responses: npt.NDArray[np.double],
        ensemble_mask: npt.NDArray[np.bool_],
        step_length: float = 0.5,
    ) -> npt.NDArray[np.double]:
        """Return a proposal for W_i, without updating the internal W.

        This is an implementation of lines 4-8 in Algorithm 1.

        Parameters
        ----------
        responses : npt.NDArray[np.double]
            The model evaluated at X_i. In other words, responses = g(X_i).
            This is Y in the paper.
        step_length : float, optional
            Step length for Gauss-Newton. The default is 0.5.

        Returns
        -------
        W_i : npt.NDArray[np.double]
            A proposal for a new W in the algorithm.

        """
        Y = responses
        assert Y.ndim == 2
        assert Y.shape[0] == self.C_dd.shape[0]
        g_X = Y.copy()
        assert ensemble_mask.dtype == bool
        assert Y.shape[1] == ensemble_mask.sum()

        # Get shapes. Same notation as used in the paper.
        N = self.X.shape[1]  # Ensemble members in prior
        n = self.X.shape[0]  # Parameters (inputs)
        m = self.C_dd.shape[0]  # Responses (outputs)
        k = Y.shape[1]  # Active ensemble members

        # Line 4 in Algorithm 1.
        Y = (g_X - g_X.mean(axis=1, keepdims=True)) / np.sqrt(N - 1)

        # Line 5
        Omega = self.W.copy()
        Omega -= Omega.mean(axis=1, keepdims=True)
        Omega /= np.sqrt(N - 1)
        np.fill_diagonal(Omega, Omega.diagonal() + 1)  # Add identity in place
        Omega = Omega[:, ensemble_mask]

        # Remind ourselves of shapes
        assert Omega.shape == (N, k)
        assert Y.shape == (m, k)

        # Line 6
        if n >= k - 1:  # Section (2.4.1)
            # Omega is not longer square, so we use least squares solver
            ST, *_ = sp.linalg.lstsq(Omega.T, Y.T)
            S = ST.T

        else:  # Section (2.4.3)
            # An alternative approach to producing A_i would be keeping the
            # returned value from the previous Newton iteration (call it X_i),
            # then computing:
            # A_i = (self.X_i - self.X_i.mean(axis=1, keepdims=True)) / np.sqrt(N - 1)
            A_i = self.A @ Omega
            ST, *_ = sp.linalg.lstsq(
                Omega.T, np.linalg.multi_dot([Y, sp.linalg.pinv(A_i), A_i]).T
            )
            S = ST.T

        # Line 7
        H = S @ self.W[:, ensemble_mask] + self.D[:, ensemble_mask] - g_X

        # Line 8
        # Some asserts to remind us what the shapes of the matrices are
        assert self.W.shape == (N, N)
        assert S.shape == (m, N)
        assert self.C_dd.shape in [(m, m), (m,)]
        assert H.shape == (m, k)

        proposed_W: npt.NDArray[np.double] = self.inversion(
            W=self.W[:, ensemble_mask],
            step_length=step_length,
            S=S,
            C_dd=self.C_dd,
            H=H,
            C_dd_cholesky=self.C_dd_cholesky,
            truncation=self.truncation,
        )
        return proposed_W


if __name__ == "__main__":
    import pytest

    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=[__file__, "--doctest-modules", "-v", "-v", "--maxfail=1"])



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.5
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %%
# ruff: noqa: E402
# %% [markdown]
# # Estimating parameters of an anharmonic oscillator
#
# The anharnomic oscillator can be modelled by a non-linear partial differential
# equation as described in section 6.3.4 of the book [Fundamentals of Algorithms
# and Data Assimilation](https://www.amazon.com/Data-Assimilation-Methods-Algorithms-Applications/dp/1611974534)
# by Mark Asch, Marc Bocquet and Maëlle Nodet.
#
# -------------
#
# The discrete two-parameter model is:
#
# $$x_{k+1} - 2 x_{k} + x_{k-1} = \omega^2 x_{k} - \lambda^2 x_{k}^3$$
#
# with initial conditions $x_0 = 0$ and $x_1 = 1$.
#
# In other words we have the following recurrence relationship:
#
# $x_{k+1} = \omega^2 x_{k} - \lambda^2 x_{k}^3 + 2 x_{k} - x_{k-1}$
#
#
# -------------
# The state vector can be written as:
#
# $$
# \mathbf{u}_k = \begin{bmatrix}
# x_{k} \\
# x_{k-1}
# \end{bmatrix}
# \quad
# \mathcal{M}_{k+1} =
#  \begin{bmatrix}
# 2 + \omega^2 - \lambda^2 x_k^2 & -1 \\
# 1 & 0
# \end{bmatrix}
# $$
#
# so that $\mathbf{u}_{k+1} = \mathcal{M}_{k+1}(\mathbf{u}_k)$.
#
#

# %%
import numpy as np
from matplotlib import pyplot as plt
from scipy import stats

import iterative_ensemble_smoother as ies

rng = np.random.default_rng(12345)

plt.rcParams["figure.figsize"] = [8, 3]


# %%
def plot_result(A, response_x_axis, title=None):
    """Plot the anharmonic oscillator, as well as marginal
    and joint distributions for the parameters."""

    responses = forward_model(A, response_x_axis)

    fig, axes = plt.subplots(1, 2 + len(A), figsize=(9, 2.25))
    if title:
        fig.suptitle(title)

    axes[0].plot(response_x_axis, responses, color="black", alpha=0.1)
    for ax, param, label in zip(axes[1:], A, ["omega", "lambda"]):
        ax.hist(param, bins="fd")
        ax.set_xlabel(label)

    axes[-1].scatter(A[0, :], A[1, :], s=5)

    fig.tight_layout()
    plt.show()


# %% [markdown]
# ## Setup


# %%
def generate_observations(K):
    """Run the model with true parameters, then generate observations."""
    # Evaluate using true parameter values on a fine grid with K sample points
    x = simulate_anharmonic(omega=3.5e-2, lmbda=3e-4, K=K)

    # Generate observations every `nobs` steps
    nobs = 50
    # Generate observation points [0, 50, 100, 150, ...] when `nobs` = 50
    observation_x_axis = np.arange(K // nobs) * nobs

    # Generate noisy observations, with value at least 5
    observation_errors = np.maximum(5, 0.2 * np.abs(x[observation_x_axis]))
    observation_values = rng.normal(loc=x[observation_x_axis], scale=observation_errors)

    return observation_values, observation_errors, observation_x_axis


def simulate_anharmonic(omega, lmbda, K):
    """Evaluate the anharmonic oscillator."""
    x = np.zeros(K)
    x[0] = 0
    x[1] = 1

    # Looping from 2 because we have initial conditions at k=0 and k=1.
    for k in range(2, K):
        x[k] = x[k - 1] * (2 + omega**2 - lmbda**2 * x[k - 1] ** 2) - x[k - 2]

    return x


def forward_model(A, response_x_axis):
    """Evaluate on each column (ensemble realization)."""
    return np.vstack(
        [simulate_anharmonic(*params, K=len(response_x_axis)) for params in A.T]
    ).T


response_x_axis = range(2500)
observation_values, observation_errors, observation_x_axis = generate_observations(
    len(response_x_axis)
)


# %% [markdown]
# ## Plot observations

# %%
fig, ax = plt.subplots(figsize=(8, 3))
ax.set_title("Anharmonic Oscillator")
ax.plot(
    np.arange(2500),
    simulate_anharmonic(omega=3.5e-2, lmbda=3e-4, K=2500),
    label="Truth",
    color="black",
)

ax.scatter(observation_x_axis, observation_values, label="Observations")

fig.tight_layout()
plt.legend()
plt.show()

# %% [markdown]
# ## Create and plot prior
#
#
# This section shows a "trick" for working with non-normal priors.
#
# Let $g$ be the forward model, so that $y = g(x)$.
# Assume for instance that $x$ must be a positive parameter.
# If so, then we can for instance use an exponential prior on $x$.
# But if we use samples from an exponential prior directly in an ensemble smoother,
# there is no guarantee that the posterior samples will be positive.
#
# The trick is to sample $x \sim \mathcal{N}(0, 1)$,
# then define a function $f$ that maps from standard normal to the
# exponential distribution.
# This function can be constructed by first mapping from standard normal to
# the interval $[0, 1)$ using the CDF, then mapping to exponential using the
# quantile function (inverse CDF) of the exponential distribution.
#
# In summary:
#
# - We send $x \sim \mathcal{N}(0, 1)$ into the ensemble smoother as before
# - We use the composite function $g(f(x))$ as our forward model, instead of $g(x)$
# - When we plot the prior and posterior, we plot $f(x)$ instead of $x$ directly

# %%
# Create Uniform prior distributions
realizations = 100
PRIORS = [stats.uniform(loc=0.025, scale=0.02), stats.uniform(loc=0.0002, scale=0.0002)]


# %%
def prior_to_standard_normal(A):
    """Map prior to U(0, 1), then use inverse of CDF to map to N(0, 1)."""
    return np.vstack(
        [stats.norm().ppf(prior.cdf(A_param)) for (A_param, prior) in zip(A, PRIORS)]
    )


def standard_normal_to_prior(A):
    """Reverse mapping."""
    return np.vstack(
        [prior.ppf(stats.norm().cdf(A_param)) for (A_param, prior) in zip(A, PRIORS)]
    )


# verify that mappings are invertible
A_uniform = np.vstack([prior.rvs(realizations) for prior in PRIORS])
assert np.allclose(
    standard_normal_to_prior(prior_to_standard_normal(A_uniform)), A_uniform
)
assert np.allclose(
    prior_to_standard_normal(standard_normal_to_prior(A_uniform)), A_uniform
)

# %%
# Create a standard normal prior and plot it in transformed space
A = rng.standard_normal(size=(2, realizations))
plot_result(standard_normal_to_prior(A), response_x_axis)

# %% [markdown]
# ## A single update step

# %%
plot_result(standard_normal_to_prior(A), response_x_axis)

# The forward model is composed with the transformation from N(0, 1) to uniform priors
responses_before = forward_model(standard_normal_to_prior(A), response_x_axis)
Y = responses_before[observation_x_axis]

# Run smoother for one step
smoother = ies.SIES(
    parameters=A,
    covariance=observation_errors**2,
    observations=observation_values,
    seed=42,
)

new_A = smoother.sies_iteration(Y, step_length=1.0)


plot_result(standard_normal_to_prior(new_A), response_x_axis)

# %% [markdown]
# ## Iterative smoother

# %%
from iterative_ensemble_smoother.utils import steplength_exponential


def iterative_smoother(A):
    A_current = np.copy(A)
    iterations = 4
    smoother = ies.SIES(
        parameters=A,
        covariance=observation_errors**2,
        observations=observation_values,
        seed=42,
    )

    plot_result(
        standard_normal_to_prior(A_current),
        response_x_axis,
        title="Prior",
    )

    for iteration in range(iterations):
        responses_before = forward_model(
            standard_normal_to_prior(A_current), response_x_axis
        )

        Y = responses_before[observation_x_axis]

        A_current = smoother.sies_iteration(
            Y, step_length=steplength_exponential(iteration + 1)
        )

        plot_result(
            standard_normal_to_prior(A_current),
            response_x_axis,
            title=f"SIES iteration {iteration + 1}",
        )


iterative_smoother(A)


# %% [markdown]
# ## ES-MDA (Ensemble Smoother - Multiple Data Assimilation)

# %%
smoother = ies.ESMDA(
    # If a 1D array is passed, it is interpreted
    # as the diagonal of the covariance matrix.
    covariance=observation_errors**2,
    observations=observation_values,
    # The inflation factors used in ESMDA
    # They are scaled so that sum_i alpha_i^-1 = 1
    alpha=np.array([8, 4, 2, 1]),
    seed=42,
)


A_current = np.copy(A)

plot_result(standard_normal_to_prior(A_current), response_x_axis, title="Prior")

for iteration in range(smoother.num_assimilations()):
    # Evaluate the model
    responses_before = forward_model(
        standard_normal_to_prior(A_current), response_x_axis
    )
    Y = responses_before[observation_x_axis]

    # Assimilate data
    A_current = smoother.assimilate(A_current, Y)

    # Plot
    plot_result(
        standard_normal_to_prior(A_current),
        response_x_axis,
        title=f"ESMDA iteration {iteration + 1}",
    )



"""
Ensemble Smoother with Multiple Data Assimilation (ES-MDA)
----------------------------------------------------------



References
----------

Emerick, A.A., Reynolds, A.C. History matching time-lapse seismic data using
the ensemble Kalman filter with multiple data assimilations.
Comput Geosci 16, 639–659 (2012). https://doi.org/10.1007/s10596-012-9275-5

Alexandre A. Emerick, Albert C. Reynolds.
Ensemble smoother with multiple data assimilation.
Computers & Geosciences, Volume 55, 2013, Pages 3-15, ISSN 0098-3004,
https://doi.org/10.1016/j.cageo.2012.03.011.

https://gitlab.com/antoinecollet5/pyesmda

"""

import numpy as np
import pytest

from iterative_ensemble_smoother.esmda import ESMDA
from iterative_ensemble_smoother.esmda_inversion import empirical_cross_covariance
from iterative_ensemble_smoother.sies import SIES


class TestESMDARealizationsDying:
    @pytest.mark.parametrize("seed", list(range(9)))
    @pytest.mark.parametrize("inversion", ESMDA._inversion_methods.keys())
    def test_that_subspaces_have_full_rank_as_realizations_die(self, seed, inversion):
        """Consider the case with twice as many realizations as parameters.

        We start with a matrix

            parameters = [ identity(10) | zeros(10, 10) ] of shape (10, 20).

        Then in each iteration one realizations dies, starting on the left.
        In the prior, the first 10 realizations span a 10-dimensional space, while
        the last 10 realizations span a 0-dimensional space.

        Here we test that once the 10 first realizations have died out, the
        10 realizations that start out as zero have been updated and span
        a 10 dimensional subspace.
        """

        rng = np.random.default_rng(seed)

        # Problem size
        ensemble_size = 20
        num_params = 10
        num_responses = 100

        # Create a linear mapping g
        A = rng.normal(size=(num_responses, num_params))

        def g(X):
            return A @ X

        # Inputs and outputs
        parameters = np.zeros(shape=(num_params, ensemble_size))
        np.fill_diagonal(parameters, 1.0)  # Identity

        x_true = np.arange(num_params)
        observations = A @ x_true + rng.normal(size=(num_responses))
        covariance = np.ones(num_responses)

        # Create smoother
        smoother = ESMDA(
            covariance=covariance,
            observations=observations,
            alpha=10,  # Number of iterations
            seed=rng,
            inversion=inversion,
        )

        # Initially, the first 10 realization span a 10 dimensional space
        assert np.linalg.matrix_rank(parameters[:, :10]) == 10

        # Initially, the last 10 realization span a 0 dimensional space
        assert np.linalg.matrix_rank(parameters[:, 10:]) == 0

        X_i = np.copy(parameters)
        ensemble_mask = np.ones(ensemble_size, dtype=bool)

        for iteration in range(smoother.num_assimilations()):
            ensemble_mask[:iteration] = False

            # Run model forward and assimilate on living realizations
            Y_i = g(X_i)
            X_i[:, ensemble_mask] = smoother.assimilate(
                X_i[:, ensemble_mask], Y_i[:, ensemble_mask]
            )

        # In the posterior, the first 10 realization should span a 10 dimensional space
        assert np.linalg.matrix_rank(X_i[:, :10]) == 10
        # In the posterior, the last 10 realization should span a 10 dimensional space
        assert np.linalg.matrix_rank(X_i[:, 10:]) == 10


class TestESMDA:
    @pytest.mark.parametrize("num_inputs", [10, 25, 50])
    @pytest.mark.parametrize("num_outputs", [5, 25, 50])
    @pytest.mark.parametrize("sies_inversion", ["direct", "subspace_exact"])
    @pytest.mark.parametrize("seed", list(range(10)))
    def test_that_ESMDA_and_SIES_produce_same_result_with_one_step(
        self, seed, sies_inversion, num_outputs, num_inputs
    ):
        """With a single step (alpha=1), ESMDA = SIES = ES.

        When num_inputs < num_ensemble - 1, then Section (2.4.3) in the SIES
        paper triggers and the result is not identical.
        """
        rng = np.random.default_rng(seed)

        num_ensemble = 10
        alpha = 1

        # Create problem instance
        X = rng.normal(size=(num_inputs, num_ensemble))
        Y = rng.normal(size=(num_outputs, num_ensemble))
        covariance = np.exp(rng.normal(size=num_outputs))
        observations = rng.normal(size=num_outputs, loc=1)

        # Create ESMDA instance and perform one iteration
        esmda = ESMDA(
            covariance, observations, alpha=alpha, seed=seed + 99, inversion="exact"
        )
        X_ESMDA = np.copy(X)

        # Perform one iteration of ESMDA
        X_ESMDA = esmda.assimilate(X_ESMDA, Y)

        # Create SIES instance and perform one iteration
        sies = SIES(
            parameters=X,
            covariance=covariance,
            observations=observations,
            inversion=sies_inversion,
            truncation=1.0,
            seed=seed + 99,
        )

        X_SIES = sies.sies_iteration(responses=Y, step_length=1)

        assert np.allclose(X_ESMDA, X_SIES)

    @pytest.mark.parametrize("seed", list(range(10)))
    @pytest.mark.parametrize("inversion", ESMDA._inversion_methods.keys())
    def test_that_diagonal_covariance_gives_same_answer_as_dense(self, seed, inversion):
        rng = np.random.default_rng(seed)

        num_outputs = rng.choice([5, 10, 15, 25])
        num_inputs = rng.choice([5, 10, 15, 25])
        num_ensemble = rng.choice([5, 10, 15, 25])
        alpha = rng.choice([1, 3, 5])

        # Prior is N(0, 1)
        X_prior = rng.normal(size=(num_inputs, num_ensemble))
        Y_prior = rng.normal(size=(num_outputs, num_ensemble))

        # Measurement errors
        covariance = np.exp(rng.normal(size=num_outputs))

        # Observations
        observations = rng.normal(size=num_outputs, loc=1)

        esmda = ESMDA(
            covariance, observations, alpha=alpha, seed=seed, inversion=inversion
        )
        X_posterior1 = np.copy(X_prior)
        for _ in range(esmda.num_assimilations()):
            X_posterior1 = esmda.assimilate(X_posterior1, Y_prior)

        esmda = ESMDA(
            np.diag(covariance),
            observations,
            alpha=alpha,
            seed=seed,
            inversion=inversion,
        )
        X_posterior2 = np.copy(X_prior)
        for _ in range(esmda.num_assimilations()):
            X_posterior2 = esmda.assimilate(X_posterior2, Y_prior)

        assert np.allclose(X_posterior1, X_posterior2)

    @pytest.mark.parametrize(
        "num_ensemble",
        [10, 100, 1000],
    )
    def test_that_alpha_as_integer_and_array_returns_same_result(self, num_ensemble):
        seed = num_ensemble
        rng = np.random.default_rng(seed)

        num_outputs = 2
        num_inputs = 1

        def g(x):
            """Transform a single ensemble member."""
            return np.array([np.sin(x / 2), x]) + 5

        def G(X):
            """Transform all ensemble members."""
            return np.array([g(x_i) for x_i in X.T]).squeeze().T

        # Prior is N(0, 1)
        X_prior = rng.normal(size=(num_inputs, num_ensemble))

        # Measurement errors
        covariance = np.eye(num_outputs)

        # The true inputs and observations, a result of running with N(1, 1)
        X_true = rng.normal(size=(num_inputs,)) + 1
        observations = G(X_true)

        # Create ESMDA instance from an integer `alpha` and run it
        esmda_integer = ESMDA(covariance, observations, alpha=5, seed=seed)
        X_i_int = np.copy(X_prior)
        for _ in range(esmda_integer.num_assimilations()):
            X_i_int = esmda_integer.assimilate(X_i_int, G(X_i_int))

        # Create another ESMDA instance from a vector `alpha` and run it
        esmda_array = ESMDA(covariance, observations, alpha=np.ones(5), seed=seed)
        X_i_array = np.copy(X_prior)
        for _ in range(esmda_array.num_assimilations()):
            X_i_array = esmda_array.assimilate(X_i_array, G(X_i_array))

        # Exactly the same result with equal seeds
        assert np.allclose(X_i_int, X_i_array)

    # Linear transformation a x + b
    @pytest.mark.parametrize("a", [-2, 1, 3, 9])
    @pytest.mark.parametrize("b", [-2, 0, 1])
    # Likelihood information
    @pytest.mark.parametrize("X_true", [-2, 5, 10, 15])
    @pytest.mark.parametrize("C_D", [1, 3, 9])
    # Prior information
    @pytest.mark.parametrize("mu", [-1, 0, 6, 12])
    @pytest.mark.parametrize("S", [1, 3, 9])
    def test_that_result_corresponds_with_theory_for_1D_gauss_linear_case(
        self, mu, S, X_true, C_D, a, b
    ):
        # Here we test on a Guass-linear case.
        # The section "2.3.3 Bayes’ theorem for Gaussian variables" in the book
        # Pattern Recognition and Machine Learning by Bishop (2006) states that if
        # p(x)   = N(x | mu, S)
        # p(y|x) = N(y | a * x + b, C_D)
        # then
        # p(x | y) has
        # covariance COV = inv(inv(S) + a * inv(C_D) * a)
        # mean      MEAN = COV (a * inv(C_D) * (y - b) + inv(S) * mu)
        def G(X):
            """Transform all ensemble members."""
            return a * X + b

        # Analytical solution given by Bishop
        def inv(x):
            # Matrix inversion for 1D matrix
            return 1 / x

        COV = inv(inv(S) + a * inv(C_D) * a)
        MEAN = COV * ((a * inv(C_D)) * (G(X_true) - b) + inv(S) * mu)

        num_ensemble = 1_000
        num_inputs = 1

        # Create a random number generator
        parameters = [mu, S, X_true, C_D, a, b]
        seed = abs(sum(p_i * 10**i for i, p_i in enumerate(parameters)))
        rng = np.random.default_rng(seed)

        # Prior is p(x) ~ N(mu, S)
        X_prior = rng.normal(size=(num_inputs, num_ensemble), loc=mu, scale=np.sqrt(S))

        # Create ESMDA instance
        esmda = ESMDA(np.atleast_2d([C_D]), G(np.atleast_1d(X_true)), alpha=1, seed=rng)
        X_i = X_prior
        for _ in range(esmda.num_assimilations()):
            X_i = esmda.assimilate(X_i, G(X_i))

        # Check that analytical solution is close to ESMDA posterior
        # np.isclose(a, b) := abs(`a` - `b`) <= (`atol` + `rtol` * abs(`b`))
        assert np.isclose(X_i.mean(), MEAN, rtol=0.1, atol=0.2)
        assert np.isclose(X_i.var(ddof=1), COV, rtol=0.1, atol=0.1)

    @pytest.mark.parametrize("seed", list(range(100)))
    def test_that_result_corresponds_with_theory_for_gauss_linear_case(self, seed):
        # Here we test on a Guass-linear case.
        # The section "2.3.3 Bayes’ theorem for Gaussian variables" in the book
        # Pattern Recognition and Machine Learning by Bishop (2006) states that if
        # p(x)   = N(x | mu, C_M)
        # p(y|x) = N(y | A x + b, C_D)
        # then
        # p(x | y) has
        # covariance COV = inv(inv(C_M) + A^T @ inv(C_D) @ A)
        # mean      MEAN = COV (A^T * inv(C_D) * (y - b) + inv(C_M) * mu)
        #                = mu + C_M A^T (A C_M A^T + C_D)^-1 A (x - mu)

        # Generate data
        rng = np.random.default_rng(seed)
        num_ensemble = 10_000
        num_inputs, num_outputs = 5, 4

        # Inputs
        mu = rng.normal(size=num_inputs)
        C_M_factor = rng.normal(size=(num_inputs, num_inputs))
        C_M = C_M_factor.T @ C_M_factor + np.eye(num_inputs)

        # Transformation
        A = rng.normal(size=(num_outputs, num_inputs))
        b = rng.normal(size=num_outputs)

        def G(x):
            if x.ndim == 1:
                return A @ x + b
            else:
                return (b.reshape(-1, 1) + (A @ x)).squeeze()

        assert np.allclose(
            G(np.arange(num_inputs)), G(np.arange(num_inputs).reshape(-1, 1)).squeeze()
        )

        # Output covariance
        C_D_factor = rng.normal(size=(num_outputs, num_outputs))
        C_D = C_D_factor.T @ C_D_factor + np.eye(num_outputs)

        # Analytical solution given by Bishop
        X_true = mu + 10
        inv = np.linalg.inv
        COV = inv(inv(C_M) + A.T @ inv(C_D) @ A)
        MEAN = COV @ ((A.T @ inv(C_D)) @ (G(X_true) - b) + inv(C_M) @ mu)
        MEAN2 = mu + C_M @ A.T @ inv(A @ C_M @ A.T + C_D) @ A @ (X_true - mu)
        assert np.allclose(MEAN, MEAN2)  # Both ways to compute mean is equivalent

        # Prior is p(x) ~ N(mu, C_M)
        X_prior = rng.multivariate_normal(mean=mu, cov=C_M, size=num_ensemble).T
        assert G(X_prior).shape == (num_outputs, num_ensemble)

        # Create ESMDA instance
        esmda = ESMDA(C_D, G(X_true), alpha=1, seed=rng)
        X_i = np.copy(X_prior)
        for _ in range(esmda.num_assimilations()):
            X_i = esmda.assimilate(X_i, G(X_i))

        # Check that analytical solution is close to ESMDA posterior
        relative_error_mean = np.linalg.norm(X_i.mean(axis=1) - MEAN) / np.linalg.norm(
            MEAN
        )
        print("relative_error_mean", relative_error_mean)

        assert relative_error_mean < 0.05

        # TODO: The errors in the covariance matrix estimation is always pretty high
        # why is this?
        covariance = empirical_cross_covariance(X_i, X_i)
        relative_error_covariance = np.linalg.norm((covariance - COV)) / np.linalg.norm(
            COV
        )
        print("relative_error_covariance", relative_error_covariance)

        assert relative_error_covariance < 1.2

    # Likelihood information
    @pytest.mark.parametrize("X_true", [5, 10])
    @pytest.mark.parametrize("C_D", [1, 3])
    # Prior information
    @pytest.mark.parametrize("mu", [0, 2, 4])
    @pytest.mark.parametrize("S", [1, 3, 9])
    # Number of iterations
    @pytest.mark.parametrize("alpha", [5, 10, 15])
    def test_that_single_and_multiple_assimil_give_same_res_for_1D_gauss_linear_case(
        self, alpha, mu, S, X_true, C_D
    ):
        # Here we test on a Guass-linear case.
        # The section "2.3.3 Bayes' theorem for Gaussian variables" in the book
        # Pattern Recognition and Machine Learning by Bishop (2006) states that if
        # p(x)   = N(x | mu, S)
        # p(y|x) = N(y | a * x + b, C_D)
        # then
        # p(x | y) has
        # covariance COV = inv(inv(S) + a * inv(C_D) * a)
        # mean      MEAN = COV (a * inv(C_D) * (y - b) + inv(S) * mu)

        def G(X):
            """Transform all ensemble members."""
            a, b = 1, 1
            return a * X + b

        num_ensemble = 10_000
        num_inputs = 1

        # Create a random number generator
        parameters = [alpha, mu, S, X_true, C_D]
        seed = abs(sum(p_i * 10**i for i, p_i in enumerate(parameters)))
        rng = np.random.default_rng(seed)

        # Prior is p(x) ~ N(mu, S)
        X_prior = rng.normal(size=(num_inputs, num_ensemble), loc=mu, scale=np.sqrt(S))

        # Create ESMDA instance with single iteration
        esmda = ESMDA(np.diag([C_D]), G(np.atleast_1d(X_true)), alpha=1, seed=rng)
        X_i_single = np.copy(X_prior)
        for _ in range(esmda.num_assimilations()):
            X_i_single = esmda.assimilate(X_i_single, G(X_i_single))

        # Create ESMDA instance with multiple iterations
        esmda = ESMDA(np.diag([C_D]), G(np.atleast_1d(X_true)), alpha=alpha, seed=rng)
        X_i_multiple = np.copy(X_prior)
        for _ in range(esmda.num_assimilations()):
            X_i_multiple = esmda.assimilate(X_i_multiple, G(X_i_multiple))

        # Check that summary statistics of solutions are close to each other
        # np.isclose(a, b) := abs(`a` - `b`) <= (`atol` + `rtol` * abs(`b`))
        assert np.isclose(X_i_single.mean(), X_i_multiple.mean(), rtol=0.05)
        assert np.isclose(X_i_single.var(), X_i_multiple.var(), rtol=0.1)


class TestESMDAMemory:
    @pytest.fixture
    def setup(self):
        rng = np.random.default_rng(42)

        num_outputs = 10_000
        num_inputs = 1_000
        num_ensemble = 100

        # Prior is N(0, 1)
        X_prior = rng.normal(size=(num_inputs, num_ensemble))
        Y_prior = rng.normal(size=(num_outputs, num_ensemble))

        # Measurement errors
        covariance = np.exp(rng.normal(size=num_outputs))

        # Observations
        observations = rng.normal(size=num_outputs, loc=1)

        return X_prior, Y_prior, covariance, observations

    @pytest.mark.limit_memory("138 MB")
    def test_ESMDA_memory_usage_subspace_inversion_without_overwrite(self, setup):
        # TODO: Currently this is a regression test. Work to improve memory usage.

        X_prior, Y_prior, covariance, observations = setup

        # Create ESMDA instance from an integer `alpha` and run it
        esmda = ESMDA(covariance, observations, alpha=1, seed=1, inversion="subspace")

        for _ in range(esmda.num_assimilations()):
            esmda.assimilate(X_prior, Y_prior)

    @pytest.mark.limit_memory("129 MB")
    def test_ESMDA_memory_usage_subspace_inversion_with_overwrite(self, setup):
        # TODO: Currently this is a regression test. Work to improve memory usage.

        X_prior, Y_prior, covariance, observations = setup

        # Create ESMDA instance from an integer `alpha` and run it
        esmda = ESMDA(covariance, observations, alpha=1, seed=1, inversion="subspace")

        for _ in range(esmda.num_assimilations()):
            esmda.assimilate(X_prior, Y_prior, overwrite=True)


@pytest.mark.parametrize("inversion", ESMDA._inversion_methods.keys())
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
@pytest.mark.parametrize("diagonal", [True, False])
def test_that_float_dtypes_are_preserved(inversion, dtype, diagonal):
    """If every matrix passed is of a certain dtype, then the output
    should also be of the same dtype. 'linalg' does not support float16
    nor float128."""

    rng = np.random.default_rng(42)

    num_outputs = 20
    num_inputs = 10
    num_ensemble = 25

    # Prior is N(0, 1)
    X_prior = rng.normal(size=(num_inputs, num_ensemble))
    Y_prior = rng.normal(size=(num_outputs, num_ensemble))

    # Measurement errors
    covariance = np.exp(rng.normal(size=num_outputs))
    if not diagonal:
        covariance = np.diag(covariance)

    # Observations
    observations = rng.normal(size=num_outputs, loc=1)

    # Convert types
    X_prior = X_prior.astype(dtype)
    Y_prior = Y_prior.astype(dtype)
    covariance = covariance.astype(dtype)
    observations = observations.astype(dtype)

    # Create ESMDA instance from an integer `alpha` and run it
    esmda = ESMDA(covariance, observations, alpha=1, seed=1, inversion=inversion)

    for _ in range(esmda.num_assimilations()):
        X_posterior = esmda.assimilate(X_prior, Y_prior)

    # Check that dtype of returned array matches input dtype
    assert X_posterior.dtype == dtype


@pytest.mark.parametrize("inversion", ESMDA._inversion_methods.keys())
def test_row_by_row_assimilation(inversion):
    # Create problem instance
    rng = np.random.default_rng(42)

    num_outputs = 4
    num_inputs = 5
    num_ensemble = 3

    A = rng.normal(size=(num_outputs, num_inputs))

    def g(X):
        return A @ X

    # Prior is N(0, 1)
    X_prior = rng.normal(size=(num_inputs, num_ensemble))

    covariance = np.exp(rng.normal(size=num_outputs))
    observations = A @ np.linspace(0, 1, num=num_inputs) + rng.normal(
        size=num_outputs, scale=0.01
    )

    # =========== Use the high level API ===========
    smoother = ESMDA(
        covariance=covariance,
        observations=observations,
        alpha=2,
        inversion=inversion,
        seed=1,
    )
    X = np.copy(X_prior)
    for iteration in range(smoother.num_assimilations()):
        X = smoother.assimilate(X, g(X))

    X_posterior_highlevel_API = np.copy(X)

    # =========== Use the low-level level API ===========
    smoother = ESMDA(
        covariance=covariance,
        observations=observations,
        alpha=2,
        inversion=inversion,
        seed=1,
    )
    X = np.copy(X_prior)
    for alpha_i in smoother.alpha:
        K = smoother.compute_transition_matrix(Y=g(X), alpha=alpha_i)

        # Here we could loop over each row in X and multiply by K
        X += X @ K

    X_posterior_lowlevel_API = np.copy(X)

    assert np.allclose(X_posterior_highlevel_API, X_posterior_lowlevel_API)


@pytest.mark.parametrize("inversion", ESMDA._inversion_methods.keys())
def test_row_by_row_assimilation_order(inversion):
    """A regression test for issue #232.

    The problem was that in a SVD call we had

        sp.linalg.svd(a, overwrite_a=True, ...)

    but the array a was used later on in the function. This bug was not spotted
    because when the input array is contiguous (C-order), then no actual overwrite
    is done by the SVD. However, in F-order it does overwrite the input array,
    which does lead to the wrong values being used later on in the function.

    The fix is to carefully review all usage of 'overwrite_a' and only keep
    overwriting if the input array is not used later on, and is not part
    of the input arguments to a function.
    """
    rng = np.random.default_rng(42)

    num_outputs = 4
    num_inputs = 5
    num_ensemble = 3

    A = rng.normal(size=(num_outputs, num_inputs))

    def g(X):
        return A @ X

    X_prior = rng.normal(size=(num_inputs, num_ensemble))
    covariance = np.exp(rng.normal(size=num_outputs))
    observations = A @ np.linspace(0, 1, num=num_inputs) + rng.normal(
        size=num_outputs, scale=0.01
    )

    # We need two instances because each call to 'compute_transition_matrix'
    # increments the random number generator.
    smoother1 = ESMDA(
        covariance=covariance,
        observations=observations,
        alpha=2,
        inversion=inversion,
        seed=1,
    )

    smoother2 = ESMDA(
        covariance=covariance,
        observations=observations,
        alpha=2,
        inversion=inversion,
        seed=1,
    )

    X = np.copy(X_prior)
    alpha_i = 2.5

    # Test that no matter which order the input is in, the output is equal
    K1 = smoother1.compute_transition_matrix(
        Y=np.ascontiguousarray(g(X)), alpha=alpha_i
    )
    K2 = smoother2.compute_transition_matrix(Y=np.asfortranarray(g(X)), alpha=alpha_i)
    assert np.allclose(K1, K2)


if __name__ == "__main__":
    import pytest

    pytest.main(
        args=[
            __file__,
            "-v",
            "-k TestESMDARealizationsDying",
        ]
    )



"""
Purpose
=======

**iterative_ensemble_smoother** is an open-source, pure python,
and object-oriented library that provides
a user friendly implementation of history matching algorithms
from :cite:t:`evensen2019efficient`.

The following functionalities are directly provided on module-level.

Classes
=======

.. autosummary::
   :toctree: _autosummary

   SIES
   ESMDA

Functions
=========

.. autosummary::
   :toctree: _autosummary

    steplength_exponential

"""

try:
    from ._version import version as __version__
    from ._version import version_tuple
except ImportError:
    __version__ = "unknown version"
    version_tuple = (0, 0, "unknown version", "unknown commit")

from iterative_ensemble_smoother.esmda import ESMDA
from iterative_ensemble_smoother.sies import SIES
from iterative_ensemble_smoother.utils import steplength_exponential

__all__ = [
    "SIES",
    "ESMDA",
    "steplength_exponential",
]



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.15.2
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %%
# ruff: noqa: E402

# %% [markdown]
# # Adaptive Localization
#
# In this example we run adaptive localization
# on a sparse Gauss-linear problem.
#
# - Each response is only affected by $3$ parameters.
# - This is represented by a tridiagonal matrix $A$ in the forward model $g(x) = Ax$.
# - The problem is Gauss-Linear, so in this case ESMDA will sample
#   the true posterior when the number of ensemble members (realizations) is large.
# - The sparse correlation structure will lead to spurious correlations, in the sense
#   that a parameter and response might appear correlated when in fact they are not.

# %% [markdown]
# ## Import packages

# %%
import numpy as np
import scipy as sp
from matplotlib import pyplot as plt

from iterative_ensemble_smoother import ESMDA
from iterative_ensemble_smoother.experimental import AdaptiveESMDA

# %% [markdown]
# ## Create problem data
#
# Some settings worth experimenting with:
#
# - Decreasing `prior_std=1` will pull the posterior solution toward zero.
# - Increasing `num_ensemble` will increase the quality of the solution.
# - Increasing `num_observations / num_parameters`
#   will increase the quality of the solution.

# %%
rng = np.random.default_rng(42)

# Dimensionality of the problem
num_parameters = 100
num_observations = 50
num_ensemble = 25
prior_std = 1

# Number of iterations to use in ESMDA
alpha = 5

# %% [markdown]
# ## Create problem data - sparse tridiagonal matrix $A$

# %%
diagonal = np.ones(min(num_parameters, num_observations))

# Create a tridiagonal matrix (easiest with scipy)
A = sp.sparse.diags(
    [diagonal, diagonal, diagonal],
    offsets=[-1, 0, 1],
    shape=(num_observations, num_parameters),
    dtype=float,
).toarray()

# We add some noise that is insignificant compared to the
# actual local structure in the forward model
A = A + rng.standard_normal(size=A.shape) * 0.01

plt.title("Linear mapping $A$ in forward model $g(x) = Ax$")
plt.imshow(A)
plt.xlabel("Parameters (inputs)")
plt.ylabel("Responses (outputs)")
plt.tight_layout()
plt.show()


# %% [markdown]
# - Below we draw prior realizations $X \sim N(0, \sigma)$.
# - The true parameter values used to generate observations are in the range $[-1, 1]$.
# - As the number of realizations (ensemble members) goes to infinity,
#   the correlation between the prior and the true parameter values converges to zero.
# - The correlation is zero for a finite number of realizations too,
#   but statistical noise might induce some spurious correlations between
#   the prior and the true parameter values.
#
# **In summary the baseline correlation is zero.**
# Anything we can do to increase the correlation above zero beats the baseline,
# which is using the prior (no update).
#
# The correlation coefficient does not take into account the uncertainty
# represented in the posterior, only the mean posterior value is compared with
# the true parameter values.
# To compare distributions we could use the
# [Kullback–Leibler divergence](https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence),
# but we do not pursue this here.


# %%
def get_prior(num_parameters, num_ensemble, prior_std):
    """Sample prior from N(0, prior_std)."""
    return rng.normal(size=(num_parameters, num_ensemble)) * prior_std


def g(X):
    """Apply the forward model."""
    return A @ X


# Create observations: obs = g(x) + N(0, 1)
x_true = np.linspace(-1, 1, num=num_parameters)
observation_noise = rng.standard_normal(size=num_observations)  # N(0, 1) noise
observations = g(x_true) + observation_noise

# Initial ensemble X ~ N(0, prior_std) and diagonal covariance with ones
X = get_prior(num_parameters, num_ensemble, prior_std)

# Covariance matches the noise added to observations above
covariance = np.ones(num_observations)  # N(0, 1) covariance

# %% [markdown]
# ## Solve the maximum likelihood problem
#
# We can solve $Ax = b$, where $b$ are the observations,
# for the maximum likelihood estimate.
#
# Notice that unlike using a Ridge model,
# solving $Ax = b$ directly does not use any prior information.

# %%
x_ml, *_ = np.linalg.lstsq(A, observations, rcond=None)

plt.figure(figsize=(7, 3))
plt.scatter(np.arange(len(x_true)), x_true, label="True parameter values")
plt.scatter(np.arange(len(x_true)), x_ml, label="ML estimate (no prior)")
plt.xlabel("Parameter index")
plt.ylabel("Parameter value")
plt.grid(True, ls="--", zorder=0, alpha=0.33)
plt.legend()
plt.tight_layout()
plt.show()

# %% [markdown]
# ## Solve using ESMDA
#
# We create an `ESMDA` instance and solve the Guass-linear problem.

# %%
smoother = ESMDA(
    covariance=covariance,
    observations=observations,
    alpha=alpha,
    seed=1,
)

X_i = np.copy(X)
for i, alpha_i in enumerate(smoother.alpha, 1):
    print(
        f"ESMDA iteration {i}/{smoother.num_assimilations()}"
        + f" with inflation factor alpha_i={alpha_i}"
    )
    X_i = smoother.assimilate(X_i, Y=g(X_i))


X_posterior = np.copy(X_i)

# %% [markdown]
# ## Plot and compare solutions
#
# Compare the true parameters with both the ML estimate
# from linear regression and the posterior means obtained using `ESMDA`.

# %%
plt.figure(figsize=(7, 3))
plt.scatter(np.arange(len(x_true)), x_true, label="True parameter values")
plt.scatter(np.arange(len(x_true)), x_ml, label="ML estimate (no prior)")
plt.scatter(
    np.arange(len(x_true)), np.mean(X_posterior, axis=1), label="Posterior mean"
)
plt.xlabel("Parameter index")
plt.ylabel("Parameter value")
plt.grid(True, ls="--", zorder=0, alpha=0.33)
plt.legend()
plt.tight_layout()
plt.show()

# %% [markdown]
# ## Solve using AdaptiveESMDA
#
# We create an `AdaptiveESMDA` instance and solve the Guass-linear problem.

# %%
adaptive_smoother = AdaptiveESMDA(
    covariance=covariance,
    observations=observations,
    seed=1,
)

X_i = np.copy(X)

# Loop over alpha defined in ESMDA instance,
# the vector of inflation values alpha is then
# of the same size in AdaptiveESMDA and ESMDA,
# and it's correctly scaled
assert np.isclose(np.sum(1 / smoother.alpha), 1), "Incorrect scaling"
for i, alpha_i in enumerate(smoother.alpha, 1):
    print(
        f"AdaptiveESMDA iteration {i}/{len(smoother.alpha)}"
        + f" with inflation factor alpha_i={alpha_i}"
    )

    # Run forward model
    Y_i = g(X_i)

    # Perturb observations
    D_i = adaptive_smoother.perturb_observations(
        ensemble_size=X_i.shape[1], alpha=alpha_i
    )

    # Assimilate data
    X_i = adaptive_smoother.assimilate(X=X_i, Y=Y_i, D=D_i, alpha=alpha_i)


X_adaptive_posterior = np.copy(X_i)

# %%
plt.figure(figsize=(7, 3))
plt.scatter(np.arange(len(x_true)), x_true, label="True parameter values")
plt.scatter(np.arange(len(x_true)), x_ml, label="ML estimate (no prior)")
plt.scatter(
    np.arange(len(x_true)), np.mean(X_posterior, axis=1), label="Posterior mean"
)
plt.scatter(
    np.arange(len(x_true)),
    np.mean(X_adaptive_posterior, axis=1),
    label="Posterior mean (adaptive)",
)
plt.xlabel("Parameter index")
plt.ylabel("Parameter value")
plt.grid(True, ls="--", zorder=0, alpha=0.33)
plt.legend()
plt.tight_layout()
plt.show()

# %% [markdown]
# ## Correlations between true parameters and solution means
#
# - A more sophisticated way to measure fit might be to use [Kullback–Leibler divergence](https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence).
# - Here we simply look at the correlations of the means.

# %%
for arr, label in zip(
    [
        np.mean(X, axis=1),
        x_ml,
        np.mean(X_posterior, axis=1),
        np.mean(X_adaptive_posterior, axis=1),
    ],
    ["Prior", "ML estimate", "Posterior mean", "Posterior mean (adaptive)"],
):
    corr = sp.stats.pearsonr(x_true, arr).statistic
    print(label, corr)


# %% [markdown]
# ## Run on several ensemble sizes and seeds
#
# To get a more complete picture of how the number of realizations (ensemble size)
# affects the result, we loop over various ensemble sizes and compute posteriors.
# To capture average behavior and reduce the influence of sampling noise,
# we also loop over several seeds.
#
# Recall that there are two sources of randomness in the result:
#
# - The draw of prior realizations $X \sim N(0, \sigma)$.
# - The noise (given by the `covariance` argument) that ESMDA adds to the observations.


# %%
def corr_true(array):
    return sp.stats.pearsonr(x_true, array).statistic


ENSEMBLE_SIZES = list(range(2, 51))
NUM_SEEDS = 25

# Store average correlation coefficients
ESMDA_corrs = []
AdaptiveESMDA_corrs = []
prior_corrs = []  # Baseline comparison - no update

# Loop over increasingly large ensemble sizes
for ensemble_size in ENSEMBLE_SIZES:
    # Posteriors means for this size
    ESMDA_means = []
    AdaptiveESMDA_means = []
    prior_means = []

    # Loop over seeds used in ESMDA/AdaptiveESMDA,
    # which determine the perturbations of the observations.
    for seed in range(NUM_SEEDS):
        # Prior
        X = get_prior(num_parameters, ensemble_size, prior_std)
        prior_means.append(np.mean(X, axis=1))

        # ================ ESMDA ==============
        smoother = ESMDA(
            covariance=covariance,
            observations=observations,
            alpha=5,
            seed=seed,
        )

        X_i = np.copy(X)
        for i, alpha_i in enumerate(smoother.alpha, 1):
            X_i = smoother.assimilate(X_i, Y=g(X_i))

        ESMDA_means.append(np.mean(X_i, axis=1))

        # ============ AdaptiveESMDA ============
        adaptive_smoother = AdaptiveESMDA(
            covariance=covariance,
            observations=observations,
            seed=seed,
        )

        X_i = np.copy(X)

        for i, alpha_i in enumerate(smoother.alpha, 1):
            # Perturb observations
            D_i = adaptive_smoother.perturb_observations(
                ensemble_size=X_i.shape[1], alpha=alpha_i
            )

            # Assimilate data
            X_i = adaptive_smoother.assimilate(X=X_i, Y=g(X_i), D=D_i, alpha=alpha_i)

        AdaptiveESMDA_means.append(np.mean(X_i, axis=1))

    # Collect results for all runs of this size
    ESMDA_corr = np.mean([corr_true(arr) for arr in ESMDA_means])
    AdaptiveESMDA_corr = np.mean([corr_true(arr) for arr in AdaptiveESMDA_means])
    prior_corr = np.mean([corr_true(arr) for arr in prior_means])

    # Add to respective lists
    ESMDA_corrs.append(ESMDA_corr)
    AdaptiveESMDA_corrs.append(AdaptiveESMDA_corr)
    prior_corrs.append(prior_corr)

# %%
# Create figure title with a lot of information
title = "ESMDA vs AdaptiveESMDA on different ensemble sizes.\n"
title += (
    f"Each point represents the average of {NUM_SEEDS} runs with different seeds.\n"
)
title += f"Parameters = {num_parameters}, Observations ="
title += f" {num_observations}, len(alpha) = {len(smoother.alpha)}"


# Create the figure
plt.figure(figsize=(7, 3.5))
plt.title(title)

plt.plot(ENSEMBLE_SIZES, ESMDA_corrs, label="ESMDA")
plt.plot(ENSEMBLE_SIZES, AdaptiveESMDA_corrs, label="AdaptiveESMDA")
plt.plot(ENSEMBLE_SIZES, prior_corrs, label="Prior")

plt.xlabel("Ensemble size")
plt.ylabel("Correlation coeff of ensemble\nmeans against true parameters")

plt.grid(True)
plt.legend()
plt.tight_layout()
plt.show()



"""
iterative_ensemble_smoother documentation build configuration file.
"""

import datetime
import os
import re
import sys
from subprocess import check_output

import iterative_ensemble_smoother as ies

package_path = os.path.abspath("..")
sys.path.insert(0, package_path)

project = "iterative_ensemble_smoother"
author = "Equinor"
copyright = f"2022-{datetime.datetime.today().year}, {author}"
# The default replacements for |version| and |release|, also used in various
# other places throughout the built documents.
version = re.sub(r"\.dev.*$", r".dev", ies.__version__)
release = version

# Convert Python files to notebooks
check_output(["jupytext", "Polynomial.py", "-o", "Polynomial.ipynb"])
check_output(["jupytext", "Oscillator.py", "-o", "Oscillator.ipynb"])
check_output(["jupytext", "LinearRegression.py", "-o", "LinearRegression.ipynb"])
check_output(["jupytext", "Adaptive.py", "-o", "Adaptive.ipynb"])

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.napoleon",  # autodoc understands numpy docstrings
    "sphinx.ext.autodoc",  # Core library for html generation from docstrings
    "sphinx.ext.autosummary",  # Create neat summary tables
    "sphinx.ext.doctest",  # Test snippets in the documentation
    "sphinx.ext.viewcode",  # Add links to highlighted source code
    "sphinx.ext.intersphinx",  # Link to other projects’ documentation
    "sphinx.ext.autosectionlabel",  # Allow reference sections using its title
    # allows BibTeX citations to be inserted into documentation generated by Sphinx
    "sphinxcontrib.bibtex",
    "jupyter_sphinx",
    "numpydoc",
    "m2r2",  # compatibility with markdown
    "myst_nb",  # to include jupyter notebooks
]

# -----------------------------------------------------------------------------
# Autosummary
# -----------------------------------------------------------------------------

# autosummaries from source-files
autosummary_generate = True
# dont show __init__ docstring
autoclass_content = "class"
# sort class members
autodoc_member_order = "groupwise"
# autodoc_member_order = 'bysource'

autosummary_generate = True  # Turn on sphinx.ext.autosummary
# This is because numpydoc screew up with autosummary
# numpydoc_show_class_members=False

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = True
napoleon_include_special_with_doc = False
# napoleon_use_admonition_for_examples = False
# napoleon_use_admonition_for_notes = False
# napoleon_use_admonition_for_references = False
napoleon_use_ivar = True
# napoleon_use_param = True
# napoleon_use_rtype = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# Add any external modules you want to refer to in the docs here.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
}

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = [".rst", ".md", ".ipynb"]

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["build", "_templates/*.rst'", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -----------------------------------------------------------------------------
# HTML output
# -----------------------------------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# html_theme = "sphinx_rtd_theme"
html_theme = "pydata_sphinx_theme"

# TODO: when ies gets a nice logo
# html_logo = '_static/logo.svg'
# html_favicon = '_static/favicon.ico'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    # "google_analytics_id": "UA-140243896-1",
    "navigation_with_keys": False,
    "show_prev_next": False,
    "github_url": "https://github.com/equinor/iterative_ensemble_smoother",
    "icon_links": [
        {
            "name": "Support",
            "url": "https://github.com/equinor/iterative_ensemble_smoother/issues",
            "icon": "fa fa-comment fa-fw",
        },
        # {
        #     "name": "The Paper",
        #     "url": "https://doi.org/10.21105/joss.01450",
        #     "icon": "fa fa-file-text fa-fw",
        # },
    ],
}

# Title displayed on the html page
html_title = "Iterative Ensemble Smoother"

# -----------------------------------------------------------------------------
# Bibliography
# -----------------------------------------------------------------------------

bibtex_bibfiles = ["./refs.bib"]
bibtex_default_style = "unsrt"
bibtex_reference_style = "author_year"
suppress_warnings = ["bibtex.duplicate_citation", "autosectionlabel.*"]

numpydoc_class_members_toctree = False

# -----------------------------------------------------------------------------
# Myst configuration
# -----------------------------------------------------------------------------

# see https://myst-parser.readthedocs.io/en/latest/syntax/optional.html#syntax-mathjax
# for math parsing in jupyter notebooks
myst_enable_extensions = [
    "amsmath",
    "dollarmath",
]



import numpy as np
import pytest
import scipy as sp
from scipy.stats import linregress

import iterative_ensemble_smoother as ies
from iterative_ensemble_smoother import SIES


class TestSIESObjective:
    @pytest.mark.parametrize("seed", list(range(99)))
    @pytest.mark.parametrize("inversion", SIES.inversion_funcs.keys())
    def test_that_sies_objective_function_decreases(self, seed, inversion):
        rng = np.random.default_rng(seed)

        ensemble_size = 50
        num_params = 25
        num_obs = num_params
        X = rng.normal(size=(num_params, ensemble_size))

        C_0 = rng.normal(size=(num_obs, ensemble_size))
        C_1 = rng.normal(size=(num_obs, num_params))
        C_2 = rng.normal(size=(num_obs, num_params))

        def G(X):
            return C_0 + C_1 @ X + 0.1 * C_2 @ X**3

        covariance = 1 + np.exp(rng.normal(size=num_obs))
        observations = rng.normal(np.zeros(num_obs), covariance)

        smoother = SIES(
            parameters=X,
            covariance=covariance,
            observations=observations,
            inversion=inversion,
            seed=rng,
        )

        Y = G(X)
        objective_before = smoother.evaluate_objective(W=smoother.W, Y=Y)

        # One iteration
        X_i = smoother.sies_iteration(Y, 0.5)
        Y_i = G(X_i)
        objective_after = smoother.evaluate_objective(W=smoother.W, Y=Y_i)

        assert objective_after <= objective_before

    @pytest.mark.parametrize("seed", list(range(99)))
    def test_that_sies_objective_function_decreases_with_many_iterations(self, seed):
        rng = np.random.default_rng(seed)

        # Create a problem instance
        ensemble_size = 5
        num_params = 10
        num_obs = num_params
        X = rng.normal(size=(num_params, ensemble_size))

        C_0 = rng.normal(size=(num_obs, ensemble_size))
        C_1 = rng.normal(size=(num_obs, num_params))
        C_2 = rng.normal(size=(num_obs, num_params))

        def G(X):
            # A slightly non-linear function
            return C_0 + C_1 @ X + 0.1 * C_2 @ X**3

        covariance = 1 + np.exp(rng.normal(size=num_obs))
        observations = rng.normal(np.zeros(num_obs), covariance)

        # F = rng.normal(size=(num_obs, num_obs))
        # observation_errors = F.T @ F

        smoother = SIES(
            parameters=X,
            covariance=covariance,
            observations=observations,
        )

        X_i = np.copy(X)

        # Initial evaluation
        Y_i = G(X_i)
        objective_before = smoother.evaluate_objective(W=smoother.W, Y=Y_i)

        # Iterations
        for iteration in range(9):
            # One iteration
            X_i = smoother.sies_iteration(Y_i, 0.01)
            Y_i = G(X_i)

            # Evaluate objective
            objective_after = smoother.evaluate_objective(W=smoother.W, Y=Y_i)

            # Check and update
            assert objective_after <= objective_before
            objective_before = objective_after

    def test_line_search(self):
        """This is a demo (or test) of how line search can be implemented."""
        rng = np.random.default_rng(42)

        ensemble_size = 5
        num_params = 10
        num_obs = num_params
        X = rng.normal(size=(num_params, ensemble_size))

        C_0 = rng.normal(size=(num_obs, ensemble_size))
        C_1 = rng.normal(size=(num_obs, num_params))
        C_2 = rng.normal(size=(num_obs, num_params))

        def G(X):
            return C_0 + C_1 @ X + 0.1 * C_2 @ X**3

        covariance = 1 + np.exp(rng.normal(size=num_obs))
        observations = rng.normal(np.zeros(num_obs), covariance)

        F = rng.normal(size=(num_obs, num_obs))
        covariance = F.T @ F

        smoother = SIES(
            parameters=X,
            covariance=covariance,
            observations=observations,
        )

        def line_search(smoother, g):
            """Demo implementation of line search."""
            BACKTRACK_ITERATIONS = 10
            self = smoother

            # Initial evaluation
            X_i = np.copy(self.X)
            N = self.X.shape[1]  # Ensemble members
            Y_i = g(X_i)

            # Perform a Gauss-Newton iteration
            while True:
                objective_before = self.evaluate_objective(W=self.W, Y=Y_i)

                # Perform a line-search iteration
                for p in range(BACKTRACK_ITERATIONS):
                    step_length = pow(1 / 2, p)  # 1, 0.5, 0.25, 0.125, ...
                    print(f"Step length: {step_length}")

                    # Make a step with the given step length
                    proposed_W = self.propose_W(Y_i, step_length)

                    # Evaluate at the new point
                    proposed_X = self.X + self.X @ proposed_W / np.sqrt(N - 1)
                    proposed_Y = g(proposed_X)
                    objective = self.evaluate_objective(W=proposed_W, Y=proposed_Y)

                    # Accept the step
                    if objective <= objective_before:
                        print(f"Accepting. {objective} <= {objective_before}")
                        # Update variables
                        self.W = proposed_W
                        X_i = proposed_X
                        Y_i = proposed_Y
                        break
                    else:
                        print(f"Rejecting. {objective} > {objective_before}")

                # If no break was triggered in the for loop, we never accepted
                else:
                    msg = (
                        "Terminating. No improvement"
                        + f"after {BACKTRACK_ITERATIONS} iterations."
                    )
                    print(msg)
                    return X_i

                yield X_i

        # Initial evaluation
        objective_before = smoother.evaluate_objective(W=smoother.W, Y=G(smoother.X))

        for X_i in line_search(smoother, G):
            objective = smoother.evaluate_objective(W=smoother.W, Y=G(X_i))
            print(objective_before, objective)


class TestSIESNumerics:
    def test_that_large_responses_are_handled(self):
        # See: https://github.com/equinor/iterative_ensemble_smoother/issues/83
        # Creating response matrix with large outlier that will
        # lead to NaNs.
        parameters = np.array([[1, 2, 3]], dtype=float)
        responses = np.array([[1, 1, 1e12], [1, 10, 100]], dtype=float)
        covariance = np.array([1, 2], dtype=float)
        observations = np.array([10, 20], dtype=float)

        # ============== SUBSPACE EXACT (RAISES) =================

        smoother = SIES(
            parameters=parameters,
            covariance=covariance,
            observations=observations,
            inversion="subspace_exact",
        )

        # Exact inversion does not work
        with pytest.raises(
            np.linalg.LinAlgError,
            match="Matrix is singular.",
        ):
            smoother.sies_iteration(responses, step_length=1.0)

        # ================== DIRECT (WARNS) =====================
        smoother = SIES(
            parameters=parameters,
            covariance=covariance,
            observations=observations,
            inversion="direct",
        )

        with pytest.warns(sp.linalg.LinAlgWarning, match="Ill-conditioned matrix"):
            smoother.sies_iteration(responses, step_length=1.0)

        # ============ SUBSPACE PROJECTED ===============
        smoother = SIES(
            parameters=parameters,
            covariance=covariance,
            observations=observations,
            inversion="subspace_projected",
        )

        # Exact inversion does not work
        smoother.sies_iteration(responses, step_length=1.0)


def test_version_attribute() -> None:
    assert ies.__version__ != "unknown version"
    assert ies.version_tuple != (0, 0, "unknown version", "unknown commit")


class TestSIESTheory:
    @pytest.mark.parametrize("number_of_realizations", [300, 500])
    def test_that_es_update_for_a_linear_model_follows_theory(
        self, number_of_realizations
    ):
        """This test:
        - Tests that the posterior covariance is
          between the prior and maximum likelihood
        - Tests that as we increase belief in the observations,
          we move closer to the maximum likelihood and further from the prior
        """
        rng = np.random.default_rng(42)

        # The following tests follow the
        # posterior properties described in
        # https://ert.readthedocs.io/en/latest/theory/ensemble_based_methods.html#kalman-posterior-properties
        a_true = 1.0
        b_true = 5.0
        number_of_observations = 45

        class LinearModel:
            rng = np.random.default_rng(42)

            def __init__(self, a, b):
                self.a = a
                self.b = b

            @classmethod
            def random(cls):
                a_std = 2.0
                b_std = 2.0
                # Priors with bias
                a_bias = 0.5 * a_std
                b_bias = -0.5 * b_std

                return cls(
                    cls.rng.normal(a_true + a_bias, a_std),
                    cls.rng.normal(b_true + b_bias, b_std),
                )

            def eval(self, x):
                return self.a * x + self.b

        true_model = LinearModel(a_true, b_true)

        ensemble = [LinearModel.random() for _ in range(number_of_realizations)]

        A = np.array(
            [
                [realization.a for realization in ensemble],
                [realization.b for realization in ensemble],
            ]
        )
        mean_prior = np.mean(A, axis=1)

        # We use time as the x-axis and observations are at
        # t=0,1,2...number_of_observations
        times = np.arange(number_of_observations)

        S = np.array([[realization.eval(t) for realization in ensemble] for t in times])

        # When observations != true model, then ml estimates != true parameters.
        # This gives both a more advanced and realistic test. Standard normal
        # N(0,1) noise is added to obtain this. The randomness ensures we are not
        # gaming the test. But the difference could in principle be any non-zero
        # scalar.
        observations = np.array(
            [true_model.eval(t) + rng.standard_normal() for t in times]
        )

        # Leading to fixed Maximum likelihood estimate.
        # It will equal true values when observations are sampled without noise.
        # It will also stay the same over beliefs.
        results = linregress(times, observations)
        maximum_likelihood = np.array([results.slope, results.intercept])

        previous_mean_posterior = mean_prior

        # numerical precision tolerance
        epsilon = 1e-2

        # We iterate with an increased belief in the observations
        for error in [10000.0, 1000.0, 100.0, 10.0, 1.0, 0.1, 0.01]:
            # An important point here is that we do not iteratively
            # update A, but instead, observations stay the same and
            # we increase our belief in the observations
            # As A is update inplace, we have to reset it.
            A = np.array(
                [
                    [realization.a for realization in ensemble],
                    [realization.b for realization in ensemble],
                ]
            )

            smoother = ies.SIES(
                parameters=A,
                covariance=np.full(observations.shape, error) ** 2,
                observations=observations,
            )

            A_posterior = smoother.sies_iteration(S, step_length=1.0)

            mean_posterior = np.mean(A_posterior, axis=1)

            # All posterior estimates lie between prior and maximum likelihood estimate
            assert (
                np.linalg.norm(mean_posterior - maximum_likelihood)
                - np.linalg.norm(mean_prior - maximum_likelihood)
                < epsilon
            )
            assert (
                np.linalg.norm(mean_prior - mean_posterior)
                - np.linalg.norm(mean_prior - maximum_likelihood)
                < epsilon
            )

            # Posterior parameter estimates improve with increased trust in observations
            assert (
                np.linalg.norm(mean_posterior - maximum_likelihood)
                - np.linalg.norm(previous_mean_posterior - maximum_likelihood)
                < epsilon
            )

            previous_mean_posterior = mean_posterior

        # At strong beliefs, we should be close to the maximum likelihood estimate
        assert np.all(
            np.linalg.norm(previous_mean_posterior - maximum_likelihood) < epsilon
        )

    @pytest.mark.parametrize("seed", list(range(9)))
    @pytest.mark.parametrize("ensemble_size", [100, 200])
    def test_that_posterior_is_between_prior_and_maximum_likelihood(
        self, ensemble_size, seed
    ):
        rng = np.random.default_rng(seed)

        # Problem size
        num_params = 10
        num_responses = 100

        # Create a linear mapping g
        A = rng.normal(size=(num_responses, num_params))

        def g(X):
            return A @ X

        # Inputs and outputs
        # The prior is given by N(0, 1)
        # The maximum likelihood estimate around N(10, 1)

        prior = rng.normal(size=(num_params, ensemble_size))
        x_true = rng.normal(size=num_params) + 10
        observations = A @ x_true + rng.normal(size=(num_responses))

        # Compute the maximum likelihood estimate using linear regression
        x_max_likelihood, *_ = np.linalg.lstsq(A, observations, rcond=None)

        # Iterate over increased belief in observations
        distance_posterior_ml_previous = 1e6
        for standard_deviation in [1e2, 1e1, 1e0, 1e-1, 1e-2]:
            covariance = np.ones(num_responses) * standard_deviation**2

            # Create smoother
            smoother = ies.SIES(
                parameters=prior,
                covariance=covariance,
                observations=observations,
                seed=rng,
            )

            # Run through the model
            responses = g(prior)

            # Only the living simulations get passed
            posterior = smoother.sies_iteration(responses, step_length=0.5)

            posterior_mean = posterior.mean(axis=1)
            prior_mean = prior.mean(axis=1)

            # The posterior should be between the prior and the ML estimate
            #     prior <----- d1 -----> posterior <------ d2 ------> ML
            #           <------------------- d3 -------------------->
            # In other words, for distances d1, d2 and d3: d3 >= d1 and d2 >= d1

            distance_prior_ml = np.linalg.norm(prior_mean - x_max_likelihood)
            distance_prior_posterior = np.linalg.norm(prior_mean - posterior_mean)
            distance_posterior_ml = np.linalg.norm(posterior_mean - x_max_likelihood)

            assert distance_prior_ml >= distance_prior_posterior
            assert distance_prior_ml >= distance_posterior_ml

            # As the observation belief increases, the posterior mean should converge
            # to the maximum likelihood estimate
            assert distance_posterior_ml <= distance_posterior_ml_previous
            distance_posterior_ml_previous = distance_posterior_ml

    @pytest.mark.parametrize("seed", list(range(25)))
    @pytest.mark.parametrize("ensemble_size", [5, 25])
    def test_that_many_small_steps_equals_one_large_step_on_linear_problem(
        self, seed, ensemble_size
    ):
        rng = np.random.default_rng(seed)

        # Problem size
        num_params = 10
        num_responses = 100

        # Create a linear mapping g
        A = rng.normal(size=(num_responses, num_params))

        def g(X):
            return A @ X

        # Inputs and outputs
        parameters = rng.normal(size=(num_params, ensemble_size))
        x_true = np.linspace(-5, 5, num=num_params)
        observations = A @ x_true + rng.normal(size=num_responses)
        covariance = np.ones(num_responses)

        # Smoother - single step with full step length
        smoother = ies.SIES(
            parameters=parameters,
            covariance=covariance,
            observations=observations,
            seed=42,
        )
        X_posterior_ES = smoother.sies_iteration(g(parameters), step_length=1.0)

        # Smoother - several shorter steps
        smoother = ies.SIES(
            parameters=parameters,
            covariance=covariance,
            observations=observations,
            seed=42,
        )
        X_i = np.copy(parameters)
        for iteration in range(18):
            X_i = smoother.sies_iteration(g(X_i), step_length=0.66)

        # The posterior is the same for a linear model
        # This is not true for non-linear models
        assert np.allclose(X_posterior_ES, X_i)

    @pytest.mark.parametrize("seed", list(range(25)))
    def test_that_posterior_covariance_is_smaller_than_prior(self, seed):
        rng = np.random.default_rng(seed)
        covariance_inflation = seed + 1

        # Problem size
        num_params = 10
        num_responses = 100
        ensemble_size = 25

        # Create a linear mapping g
        A = rng.normal(size=(num_responses, num_params))

        def g(X):
            return A @ X

        # Inputs and outputs
        prior = rng.normal(size=(num_params, ensemble_size))
        x_true = np.linspace(-5, 5, num=num_params)
        observations = A @ x_true + rng.normal(size=num_responses)
        covariance = np.ones(num_responses) * covariance_inflation

        # Smoother - single step with full step length
        smoother = ies.SIES(
            parameters=prior,
            covariance=covariance,
            observations=observations,
            seed=seed,
        )
        posterior = smoother.sies_iteration(g(prior), step_length=1.0)

        # We are more certain after assimilating, so the covariance decreases
        assert np.linalg.det(np.cov(posterior)) < np.linalg.det(np.cov(prior))

        # Posterior mean is closer to true value than prior mean
        posterior_mean = posterior.mean(axis=1)
        prior_mean = prior.mean(axis=1)
        assert np.linalg.norm(posterior_mean - x_true) < np.linalg.norm(
            prior_mean - x_true
        )

    @pytest.mark.parametrize("inversion", list(ies.SIES.inversion_funcs.keys()))
    @pytest.mark.parametrize("variance", [1, 2, 5, 9])
    def test_that_update_correctly_multiples_gaussians_in_1D(self, inversion, variance):
        """Test that SIES follows the theory on the Guass-linear case.
        See section "2.3.3 Bayes’ theorem for Gaussian variables" in the book
        "Pattern Recognition and Machine Learning" by Bishop (2006) for more info.
        Also, see Section 8.1.8 in "the matrix cookbook".

        NB! This test is potentially flaky because of finite ensemble size.

        Assume p(x) is N(mu=0, Sigma=2I) and p(y|x) is N(mu=y, Sigma=2I).
        Multiplying these together, we get that p(x|y) is N(mu=y/2, Sigma=I).
        Note that Sigma is a covariance matrix.

        Here we use this property, and assume that the forward model is the identity.
        """
        N = 1000

        rng = np.random.default_rng(variance)

        # X ~ N(0, I * var)
        X = rng.normal(scale=np.sqrt(variance), size=(1, N))

        # Assuming forward model is the identity
        Y = np.copy(X)

        obs_val = 10.0
        covariance = np.ones(1) * variance
        observations = np.ones(1) * obs_val
        smoother = ies.SIES(
            parameters=X, covariance=covariance, observations=observations, seed=rng
        )

        X_posterior = smoother.sies_iteration(
            Y,
            step_length=1.0,
        )

        # Posterior mean should end up between 0 and obs_val (in the middle),
        # when the covariances are equal and the mapping is the identity.
        posterior_mean = X_posterior.mean(axis=1)
        assert np.allclose(posterior_mean, obs_val / 2, rtol=0.1)

        # Posterior covariance
        posterior_variance = 1 / (1 / variance + 1 / variance)
        assert np.allclose(np.var(X_posterior), posterior_variance, rtol=0.1)

    @pytest.mark.parametrize("inversion", list(ies.SIES.inversion_funcs.keys()))
    @pytest.mark.parametrize("variance", [0.25, 0.5, 1, 2, 4])
    def test_that_update_correctly_multiples_gaussians(self, inversion, variance):
        """Test that SIES follows the theory on the Guass-linear case.
        See section "2.3.3 Bayes’ theorem for Gaussian variables" in the book
        "Pattern Recognition and Machine Learning" by Bishop (2006) for more info.
        Also, see Section 8.1.8 in "the matrix cookbook".

        NB! This test is potentially flaky because of finite ensemble size.

        Assume p(x) is N(mu=0, Sigma=2I) and p(y|x) is N(mu=y, Sigma=2I).
        Multiplying these together, we get that p(x|y) is N(mu=y/2, Sigma=I).
        Note that Sigma is a covariance matrix.

        Here we use this property, and assume that the forward model is the identity.
        """
        N = 1000
        nparam = 3

        rng = np.random.default_rng(int(variance * 100))

        # X ~ N(0, I * var)
        X = rng.normal(scale=np.sqrt(variance), size=(nparam, N))

        # Assuming forward model is the identity
        Y = np.copy(X)

        obs_val = 10.0
        covariance = np.ones(nparam) * variance
        observations = np.ones(nparam) * obs_val
        smoother = ies.SIES(
            parameters=X, covariance=covariance, observations=observations, seed=rng
        )

        X_posterior = smoother.sies_iteration(
            Y,
            step_length=1.0,
        )

        # Posterior mean should end up in the middle
        posterior_mean = X_posterior.mean(axis=1)
        assert np.allclose(posterior_mean, obs_val / 2, rtol=0.1)

        theoretical_variance = 1 / (1 / variance + 1 / variance)
        assert np.allclose(
            np.cov(X_posterior), np.eye(nparam) * theoretical_variance, atol=0.33
        )


@pytest.mark.parametrize(
    "ensemble_size,num_params,linear",
    [
        pytest.param(100, 3, True, id="No projection because linear model"),
        pytest.param(
            100,
            3,
            False,
            id="Project because non-linear and num_params < ensemble_size - 1",
        ),
        pytest.param(3, 100, True, id="No projection because linear model"),
        pytest.param(
            3, 100, False, id="No projection because num_params > ensemble_size - 1"
        ),
        pytest.param(
            10, 10, True, id="No projection because num_params > ensemble_size - 1"
        ),
    ],
)
def test_that_global_es_update_is_identical_to_local(ensemble_size, num_params, linear):
    rng = np.random.default_rng(42)

    num_obs = num_params
    X = rng.normal(size=(num_params, ensemble_size))
    Y = X if linear else np.power(X, 2)
    covariance = np.exp(rng.uniform(size=num_obs))
    observations = rng.normal(np.zeros(num_params), covariance)

    # First update a subset or batch of parameters at once
    # and and then loop through each remaining parameter and
    # update it separately.
    # This is meant to emulate a configuration where users
    # want to update, say, a field parameter separately from scalar parameters.

    # A single global update
    smoother = ies.SIES(
        parameters=X, covariance=covariance, observations=observations, seed=1
    )
    X_ES_global = smoother.sies_iteration(Y, step_length=1.0)

    # Update each parameter in turn
    smoother = ies.SIES(
        parameters=X, covariance=covariance, observations=observations, seed=1
    )
    W = smoother.propose_W(Y, step_length=1.0)

    # Create an empty matrix, then fill it in row-by-row
    X_ES_local = np.empty(shape=(num_params, ensemble_size))
    for i in range(X.shape[0]):
        # This is line 9 in algorithm 1, using W
        X_ES_local[i, :] = X[i, :] + X[i, :] @ W / np.sqrt(ensemble_size - 1)

    assert np.allclose(X_ES_global, X_ES_local)


@pytest.mark.parametrize("seed", list(range(10)))
def test_that_ies_runs_with_failed_realizations(seed):
    """When the forward model is a simulation being run in parallel on clusters,
    some of the jobs (realizations) might fail. In that case we know which
    realizations have failed, and we wish to keep assimilating data with the
    remaining realizations.

    We assume that once a realization is "dead", it never wakes up again.
    """

    rng = np.random.default_rng(seed)

    # Problem size
    ensemble_size = 300
    num_params = 20
    num_responses = 200

    # Create a linear mapping g
    A = rng.normal(size=(num_responses, num_params))

    def g(X):
        return A @ X

    # Inputs and outputs
    parameters = rng.normal(size=(num_params, ensemble_size))
    x_true = np.linspace(-5, 5, num=num_params)
    observations = A @ x_true + rng.normal(size=(num_responses))
    covariance = np.ones(num_responses)

    # Create smoother
    smoother = ies.SIES(
        parameters=parameters,
        covariance=covariance,
        observations=observations,
        seed=rng,
    )

    ensemble_mask = np.ones(ensemble_size, dtype=bool)
    X_i = np.copy(parameters)
    for iteration in range(10):
        # Assume we do a model run, but some realizations (ensemble members)
        # fail for some reason. We know which realizations failed.
        # With 10 iterations and p=[0.9, 0.1], around 1/2 make it till the end.
        simulation_ok = rng.choice([True, False], size=ensemble_size, p=[0.95, 0.05])
        ensemble_mask = np.logical_and(ensemble_mask, simulation_ok)

        # Run through the model
        responses = g(X_i)

        # Only the living simulations get passed
        X_i = smoother.sies_iteration(
            responses[:, ensemble_mask], ensemble_mask=ensemble_mask, step_length=0.05
        )

        # The full matrix is returned, but only X_i[:, ensemble_mask] are active
        assert X_i.shape == parameters.shape

    mean_posterior_all = X_i.mean(axis=1)
    mean_posterior = X_i[:, ensemble_mask].mean(axis=1)

    # Using only the realizations that lived through all iterations is better
    # than using the ones who died along the way
    assert np.linalg.norm(mean_posterior - x_true) < np.linalg.norm(
        mean_posterior_all - x_true
    )

    # The posterior is closer to the true value than the prior
    assert np.linalg.norm(mean_posterior - x_true) < np.linalg.norm(
        parameters.mean(axis=1) - x_true
    )

    # The posterior mean is reasonably close - or at least nothing crazy happens
    assert np.all(np.abs(mean_posterior - x_true) < 3)


@pytest.mark.parametrize("seed", list(range(9)))
def test_that_subspaces_are_still_used_as_original_realizations_fail(seed):
    """When the forward model is a simulation being run parallel on clusters,
    some of the jobs (realizations) might fail.

    Consider the case when the prior is the identity matrix I.
    N = n = 10. In each iteration, another ensemble member "dies."
    After 7 iterations, only the three right-most columns are still alive.
    However, that does NOT mean that the posterior lies in the subspace
    spanned by the three most right-most columns in the prior, since
    in earlier iterations the algorithm is able to explore other subspaces
    before realizations die out.
    """

    rng = np.random.default_rng(seed)

    # Problem size
    ensemble_size = num_params = 10
    num_responses = 100

    # Create a linear mapping g
    A = rng.normal(size=(num_responses, num_params))

    def g(X):
        return A @ X

    # Inputs and outputs
    parameters = np.identity(num_params)  # Identity
    x_true = np.arange(num_params)
    observations = A @ x_true + rng.normal(size=(num_responses))
    covariance = np.ones(num_responses)

    # Create smoother
    smoother = ies.SIES(
        parameters=parameters,
        covariance=covariance,
        observations=observations,
        seed=rng,
    )

    X_i = np.copy(parameters)
    ensemble_mask = np.ones(ensemble_size, dtype=bool)
    for iteration in range(1, 8):
        ensemble_mask[:iteration] = False

        # Only the living simulations get passed
        responses = g(X_i)
        X_i = smoother.sies_iteration(
            responses[:, ensemble_mask], ensemble_mask=ensemble_mask, step_length=0.2
        )

    # First column is the identity, it was never updated
    assert X_i[0, 0] == 1
    assert np.allclose(X_i[1:, 0], 0)

    # Variables keep getting updated, index (i, j) != (i, j+k) for k > 0
    assert not np.allclose(X_i[1, 1:], X_i[1, 1])
    assert not np.allclose(X_i[2, 2:], X_i[2, 2])
    assert not np.allclose(X_i[3, 3:], X_i[3, 3])

    # If the first 7 columns were not used, and the result lied in the subspace
    # spanned by the last 3 columns, then the top 7 rows would be
    # identitally zero, but that is not the case
    assert not np.allclose(X_i[:7, ensemble_mask], 0)


@pytest.mark.parametrize("seed", list(range(9)))
def test_that_subspaces_have_full_rank_as_realizations_die(seed):
    """Consider the case with twice as many realizations as parameters.

    We start with a matrix

        parameters = [ identity(10) | zeros(10, 10) ] of shape (10, 20).

    Then in each iteration one realizations dies, starting on the left.
    In the prior, the first 10 realizations span a 10-dimensional space, while
    the last 10 realizations span a 0-dimensional space.

    Here we test that once the 10 first realizations have died out, the
    10 realizations that start out as zero have been updated and span
    a 10 dimensional subspace.
    """

    rng = np.random.default_rng(seed)

    # Problem size
    ensemble_size = 20
    num_params = 10
    num_responses = 100

    # Create a linear mapping g
    A = rng.normal(size=(num_responses, num_params))

    def g(X):
        return A @ X

    # Inputs and outputs
    parameters = np.zeros(shape=(num_params, ensemble_size))
    np.fill_diagonal(parameters, 1.0)  # Identity

    x_true = np.arange(num_params)
    observations = A @ x_true + rng.normal(size=(num_responses))
    covariance = np.ones(num_responses)

    # Create smoother
    smoother = ies.SIES(
        parameters=parameters,
        covariance=covariance,
        observations=observations,
        seed=rng,
    )

    # Initially, the first 10 realization span a 10 dimensional space
    assert np.linalg.matrix_rank(parameters[:, :10]) == 10

    # Initially, the last 10 realization span a 0 dimensional space
    assert np.linalg.matrix_rank(parameters[:, 10:]) == 0

    X_i = np.copy(parameters)
    ensemble_mask = np.ones(ensemble_size, dtype=bool)
    for iteration in range(20):
        ensemble_mask[:iteration] = False

        # Only the living simulations get passed
        responses = g(X_i)
        X_i = smoother.sies_iteration(
            responses[:, ensemble_mask], ensemble_mask=ensemble_mask, step_length=0.2
        )

    # In the posterior, the first 10 realization should span a 10 dimensional space
    assert np.linalg.matrix_rank(X_i[:, :10]) == 10
    # In the posterior, the last 10 realization should span a 10 dimensional space
    assert np.linalg.matrix_rank(X_i[:, 10:]) == 10


@pytest.mark.parametrize("ensemble_size", [5, 15, 50])
def test_that_full_ensemble_mask_is_equal_to_no_ensemble_mask(ensemble_size):
    rng = np.random.default_rng(42)

    # Problem size
    num_params = 10
    num_responses = 25

    # Create a linear mapping g
    A = rng.normal(size=(num_responses, num_params))

    def g(X):
        return A @ X

    # Inputs and outputs
    parameters = rng.normal(size=(num_params, ensemble_size))
    x_true = rng.normal(size=(num_params))
    observations = A @ x_true + rng.normal(size=(num_responses))
    covariance = np.ones(num_responses)

    X_i = np.copy(parameters)
    responses = g(X_i)
    ensemble_mask = np.ones(ensemble_size, dtype=bool)

    # Create smoother with the intention of using ensemble mask
    smoother1 = ies.SIES(
        parameters=parameters,
        covariance=covariance,
        observations=observations,
        seed=1,
    )
    assert np.allclose(responses[:, ensemble_mask], responses)
    X_i_mask = smoother1.sies_iteration(
        responses[:, ensemble_mask], ensemble_mask=ensemble_mask, step_length=0.2
    )

    # Create new smoother with no ensemble mask
    smoother2 = ies.SIES(
        parameters=parameters,
        covariance=covariance,
        observations=observations,
        seed=1,
    )
    X_i = smoother2.sies_iteration(responses, step_length=0.2)

    # Same result
    assert np.allclose(X_i_mask, X_i)


@pytest.mark.parametrize("seed", list(range(10)))
@pytest.mark.parametrize("inversion", list(ies.SIES.inversion_funcs.keys()))
def test_that_diagonal_and_dense_covariance_return_the_same_result(inversion, seed):
    """A 1D array represents the diagonal of a covariance matrix. The user should
    be able to pass either a 2D or 1D representation of a diagonal, and get the same
    results. But a 1D representation is more efficient w.r.t. speed and memory."""
    rng = np.random.default_rng(seed)

    # Create a random problem instance
    ensemble_size = rng.choice([5, 10, 50, 100])
    num_params = rng.choice([5, 10, 50, 100])
    num_obs = rng.choice([5, 10, 50, 100])

    X = rng.normal(size=(num_params, ensemble_size))  # X
    responses = rng.normal(size=(num_obs, ensemble_size))  # Y
    observations = rng.normal(size=num_obs)  # d

    covariance_1D = np.exp(rng.normal(size=num_obs))
    covariance_2D = np.diag(covariance_1D)

    # 1D array of standard deviations
    smoother_diag = ies.SIES(
        parameters=X, covariance=covariance_1D, observations=observations, seed=seed
    )
    assert covariance_1D.ndim == 1
    X_post_diag = smoother_diag.sies_iteration(responses)

    # 2D array of covariances (covariance matrix)
    smoother_covar = ies.SIES(
        parameters=X, covariance=covariance_2D, observations=observations, seed=seed
    )
    assert covariance_2D.ndim == 2
    X_post_covar = smoother_covar.sies_iteration(responses)

    assert np.allclose(X_post_diag, X_post_covar)  # Same result
    assert not np.allclose(X, X_post_covar)  # Update happened


@pytest.mark.parametrize("inversion", list(ies.SIES.inversion_funcs.keys()))
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
@pytest.mark.parametrize("diagonal", [True, False])
def test_that_float_dtypes_are_preserved(inversion, dtype, diagonal):
    """If every matrix passed is of a certain dtype, then the output
    should also be of the same dtype. 'linalg' does not support float16
    nor float128."""

    rng = np.random.default_rng(42)

    # Create a random problem instance
    ensemble_size = 25
    num_params = 10
    num_obs = 20

    parameters = rng.normal(size=(num_params, ensemble_size))  # X
    responses = rng.normal(size=(num_obs, ensemble_size))  # Y
    observations = rng.normal(size=num_obs)  # d

    covar_factor = rng.normal(size=(num_obs, num_obs))
    covariance = covar_factor.T @ covar_factor
    if diagonal:
        covariance = np.diag(covariance)

    # Convert dtypes
    parameters = parameters.astype(dtype)
    covariance = covariance.astype(dtype)
    observations = observations.astype(dtype)
    responses = responses.astype(dtype)

    smoother = ies.SIES(
        parameters=parameters,
        covariance=covariance,
        observations=observations,
        seed=rng,
        inversion=inversion,
    )

    X_posterior = smoother.sies_iteration(responses)

    assert X_posterior.dtype == dtype


@pytest.mark.limit_memory("70 MB")
def test_memory_usage():
    """Estimate expected memory usage and make sure ES does not waste memory

    # approx. 65
    # Size of input arrays
    nbytes = (
        X.nbytes
        + Y.nbytes
        + observation_errors.nbytes
        + observation_values.nbytes
        + noise.nbytes
    )
    nbytes += noise.nbytes  # Creating E
    nbytes += noise.nbytes  # Creating D
    # scaling response_ensemble
    # (can't scale in-place because response_ensemble is an input argument)
    nbytes += (
        noise.nbytes
    )
    nbytes += 80000 # Omega in C++ (ensemble_size, ensemble_size)
    nbytes += Y.nbytes # Solving for S^T needs Y^T which causes a copy in C++ code
    nbytes += Y.nbytes # Solving for S^T causes both Y^T and S^T to be in memory
    nbytes += Y.nbytes # Creating H in C++
    nbytes /= 1e6
    """
    rng = np.random.default_rng(42)

    ensemble_size = 100
    num_params = 1000
    num_obs = 10000
    X = rng.normal(size=(num_params, ensemble_size))

    Y = rng.normal(size=(num_obs, ensemble_size))

    observation_errors = rng.uniform(size=num_obs)
    observation_values = rng.normal(np.zeros(num_obs), observation_errors)

    smoother = ies.SIES(
        parameters=X,
        covariance=observation_errors,
        observations=observation_values,
    )

    smoother.sies_iteration(Y, 1.0)


if __name__ == "__main__":
    import pytest

    # --durations=10  <- May be used to show potentially slow tests

    pytest.main(
        args=[
            __file__,
            "--doctest-modules",
            "-v",
            #  "-k test_that_update_correctly_multiples_gaussians",
        ]
    )
